/**************************************************************************

	handlestrings.c: Some basic functions for manipulating text
	Copyright 2002, Jacob A. Ewing

	Author:	Jacob A. Ewing
	e-mail:	jacob@diagrami.com

**************************************************************************

    This file is part of Netcrib.

    Netcrib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Netcrib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Netcrib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************/

#ifndef NULL
	#define NULL 0
#endif


/*****
the same as strlen() in string.h, returns the number of 
characters in an array before hitting a null character.
*****/
int stringlength(char *string){
	int n;
	for(n = 0; string[n] != (char)NULL; n++);
	return n;
}

/***
findstring searches for the text that "*search" points to as a substring
of the text that "*text" points to.
Upon success:
if mode == 0, it returns the starting position of that substring.
if mode != 0, it returns the ending position of that substring.
Upon failure (ie. not finding the substring), it returns -1;
***/
int findstring(char *text, char *search, int mode){
	int n, m, tlen, slen, foundit, returnval;
	returnval = -1;

	foundit = 0;
	tlen = stringlength(text);
	slen = stringlength(search);

	for(n = 0; text[n] != (char)NULL && !foundit; n++){
		for(m = 0; (text[n + m] == search[m]) && (n + m < tlen) && (m < slen); m++);
		if(m >= slen)
			foundit = 1;
	}

	if(foundit){
	        if(!mode) returnval = n - 1;
	        else returnval = m + n - 1;
	}
	return returnval;
}

/*******
Takes an integer value and returns a string with that number
expressed in any numeric base from 2 to 36;
for intstance:
itoa(narf, 83, 2); would assign the value "1010011" to the character array "narf"
itoa(narf, 83, 30); would assign the value "2N" to the character array "narf"
itoa(narf, 83, 10); would assign the value "83" to the character array "narf"
itoa(narf, 83, 1); would leave "narf" untouched and return NULL;
*******/
char *itoa(char *dest, int val, int base){
	char *returnval = (char *)NULL;
	int digit, n, returnlen, negative = 0;

	if(base <= 36 && base > 1){
		returnlen = 0;
		dest[returnlen] = NULL;
		if(val < 0) negative = 1;
		
		do{
			returnlen ++;
			for(n = returnlen; n > 0; n--){
				dest[n] = dest[n - 1];
			}

			digit = val % base;
			
			if(digit < 10)
				dest[0] = ('0' + (char)digit);
			else{
				digit -= 10;
				dest[0] = ('A' + (char)digit);
			}
			val /= base;
		}while(val);

		if(negative){
			returnlen ++;
			for(n = returnlen; n > 0; n--){
				dest[n] = dest[n - 1];
			}
			dest[0] = '-';
		}

		returnval = dest;
	}
	
	return returnval;
}
