/*****************************************************************************
	main.cpp:	 Copyright 2002, Jacob A. Ewing
	This is the main body of the project.  I am currently in the process
	of breaking this down into smaller components.
	for more information see README

	Author:	Jacob A. Ewing
	e-mail: ev910@ncf.ca

******************************************************************************

    This file is part of Netcrib.

    Netcrib is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Netcrib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Netcrib; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*****************************************************************************/
/******** header files ********/
#include <stdarg.h>
#include <iostream.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <ncurses.h>

/****** local header files ******/
#include "handlestrings.h"
#include "boardclass.h"
#include "cardclasses.h"

/****** definitions *********/
/**
an odd number here:  MAXHANDSIZE is set to the largest possible number of 
cards that can be played before hitting 31 when pegging.  This is because
this program handles pegging by duplicating them in another hand named
"pile" and checking for points in that.
**/
#ifndef MAXHANDSIZE
	#define MAXHANDSIZE 13
#endif /*MAXHANDSIZE*/
#define maxplayers 4
#define MAXNAMELENGTH 16
#define DEFAULT_PORT_NUMBER 5555
#define DEFAULT_BOARDFILENAME "2player.board"
#define KEYSIZE 8

#define CARDBACK_COLOUR 5
#define PLAYER_TEXT_COLOUR 6
#define MESSAGE_TEXT_COLOUR 8

/****************** enumerations *******************/
#ifndef DEFDISPLAYMODETYPE
	#define DEFDISPLAYMODETYPE
	enum displaymodetype { none, ansi, opengl };
#endif
enum playmodetype{ HAND, CRIB, PEG, NOCUT };
enum playstyletype { host, client, local, nostyle};
enum playermodetype { console, remote, bot1, bot2 };

/****************** classes *******************/
class playertype{
    private:
	void divide_by_human(void);
	void divide_by_computer(void);

    public:
	int score, priorscore, myindexnumber;
	playermodetype typeofplayer;
	handtype hand;
	char name[MAXNAMELENGTH], key[KEYSIZE];

	playertype(){
		score = priorscore = myindexnumber = 0;
		typeofplayer = console;
	}

	void getname(void){
		if(typeofplayer != console || 1 /********** this "if" needs to be changed later **************/){
			switch(rand() % 26){
				case 0: strcpy(name, "Alfred"); break;
				case 1: strcpy(name, "Bonnie"); break;
				case 2: strcpy(name, "Charles"); break;
				case 3: strcpy(name, "Dominique"); break;
				case 4: strcpy(name, "Edward"); break;
				case 5: strcpy(name, "Francine"); break;
				case 6: strcpy(name, "Ginny"); break;
				case 7: strcpy(name, "Howard"); break;
				case 8: strcpy(name, "Ivan"); break;
				case 9: strcpy(name, "Josephine"); break;
				case 10: strcpy(name, "Kenny"); break;
				case 11: strcpy(name, "Laurel"); break;
				case 12: strcpy(name, "Michael"); break;
				case 13: strcpy(name, "Nora"); break;
				case 14: strcpy(name, "Oliver"); break;
				case 15: strcpy(name, "Penelope"); break;
				case 16: strcpy(name, "Quincy"); break;
				case 17: strcpy(name, "Rebecca"); break;
				case 18: strcpy(name, "Samson"); break;
				case 19: strcpy(name, "Tasha"); break;
				case 20: strcpy(name, "Uriel"); break;
				case 21: strcpy(name, "Victoria"); break;
				case 22: strcpy(name, "William"); break;
				case 23: strcpy(name, "Xena"); break;
				case 24: strcpy(name, "Yolanda"); break;
				case 25: strcpy(name, "Zachary"); break;
			}
		}
	}

	void showhand(displaymodetype displaymode){
		hand.display(displaymode);
	}

	void divide_hand(void){
		if(typeofplayer == console) divide_by_human();
		else divide_by_computer();
	}

	int play_card(handtype *pile);
};

/****************** function declarations *********************/
void randomize(void);
int handle_arguments(int argc, char *argv[]);
void deal_round(void);
int count_points(handtype hand, playmodetype playmode, displaymodetype displaymode);
int count_fifteens(handtype hand, playmodetype playmode, displaymodetype displaymode);
int count_runs(handtype hand, playmodetype playmode, displaymodetype displaymode);
int count_pairs(handtype hand, playmodetype playmode, displaymodetype displaymode);
int count_flush(handtype hand, playmodetype playmode, displaymodetype displaymode);
int count_nobs(handtype hand, playmodetype playmode, displaymodetype displaymode);
int count_play_value(handtype pile, int current_player, displaymodetype displaymode);
void divide_hands(void);
void dealfromdeck(decktype *d, playertype *p, int numtodeal);
void cut_deck(void);
void play_hands(void);
void peg_hands(displaymodetype displaymode);
void show_winner(displaymodetype displaymode);
void get_player_names(void);
void display_round_start(int numrounds);
void restock_deck(void);
void draw_cardback(int y, int x);
void draw_pegs(void);
void box(int y1, int x1, int ysize, int xsize, char value);
int load_board(char *filename);
void draw_board(boardtype *board, int starty, int startx);
void initialize_windows(void);
void redraw_screen(void);
void increment_score(int playernum, int amount);
char wait_for_keypress(void);
void initialize_screen(void);
void set_default_values(void);
char binaryprompt(char *text, char option1, char option2, int case_sensitive);
void play_local_game(void);
void get_game_settings(void);
void writemessage(char *message, ...);

/****************** global variables ******************/
playertype player[maxplayers];
decktype deck;
cardtype cutcard;
int numplayers, handsize, starting_handsize, starting_cribsize, add_to_crib_size, current_dealer;
int game_over, winning_player, maximum_score;
displaymodetype main_displaymode;
handtype crib, stack;
int current_playmode = 0, xscreensize, yscreensize, port_number;
boardtype board;
int numcolourpairs, baseboardcolour, boardx, boardy, messagex, messagey;
WINDOW *messagewindow;
playstyletype playstyle;

char host_address[1024], *boardfilename, localplayername[256];
int numclients;

/****************** function definitions *********************/
int main(int argc, char *argv[]){
	int n;
	randomize();

	set_default_values();
	
	n = handle_arguments(argc, argv);
	switch(n){
		case 0: break;
		case -1: exit(0); break;
		default: exit(n);
	}
	
	initialize_screen();				/** initialize the screen **/
	if(load_board(boardfilename)) exit(1);		/** load the board **/
	initialize_windows();				/** initialize any needed windows **/

	get_game_settings();
	
	/*** Give each of the players a number as well.  This is used
	to tell within the objects themselves just who they are
	**/
	for(n = 0; n < numplayers; n++) player[n].myindexnumber = n;
	play_local_game();
	if(main_displaymode == ansi){
		delwin(messagewindow);
		endwin();
	}

	free(boardfilename);
	
	return 0;
}

/********
This gets all the basic settings for what type of game you want to play.
********/
void get_game_settings(void){
	char c, text[60];
	int n, needportnumber = 0;


	clear();
	playstyle = local;
	player[0].typeofplayer = console;
	for(n = 1; n < numplayers; n++){
		player[n].typeofplayer = bot2;
	}
	get_player_names();
}

/*****
Play the game.
*****/
void play_local_game(void){
	int n, numrounds, play_again;
	char c;

	do{

		/******* initialize values that need to be reset for every game ********/
		for(n = 0; n < numplayers; n++){
			player[n].score = player[n].priorscore = 0;
		}
		numrounds = game_over = 0;
		if(main_displaymode == ansi){
			clear();
			draw_board(&board, boardy, boardx);
		}

		while(!game_over){
			if(main_displaymode == ansi){
				box(0, 0, yscreensize, boardx - 1, ' ');
				draw_cardback((yscreensize >> 1) - 1, 0);
			}

			numrounds ++;
			display_round_start(numrounds);
			deck.shuffle();
			deal_round();
			divide_hands();
			cut_deck();
			if(!game_over) play_hands();
			if(!game_over) peg_hands(main_displaymode);
	
			restock_deck();
	
		       	if(main_displaymode == ansi && !game_over){
				wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
				wprintw(messagewindow, "Press a key for the next round.\n");
				wrefresh(messagewindow);
				wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
				wait_for_keypress();
			}
			current_dealer = (current_dealer + 1) % numplayers;
		}
		if(main_displaymode == ansi) refresh();
		show_winner(main_displaymode);

		if(main_displaymode == ansi){
			wprintw(messagewindow, "Play again? (y/n)\n");
			wrefresh(messagewindow);
		}

		do{
			c = wgetch(messagewindow);
			if(c >= 'A' && c <= 'Z') c -= ('A' - 'a');
		}while(c != 'y' && c != 'n');


		play_again = (c == 'y');
		current_dealer = (winning_player + 1) % numplayers;
	}while(play_again);
}

/****
Assign default values to some global variables.
****/
void set_default_values(void){
	/*** set the default values ***/
	game_over = 0;					/* non-zero if the game has been won */
	main_displaymode = ansi;			/* primary mode of display */
	playstyle = nostyle;				/* default type of game */
	port_number = DEFAULT_PORT_NUMBER;		/* set the port number to use for a network game */
	boardfilename = strdup(DEFAULT_BOARDFILENAME);	/* set the name of the file from which to load the board info */
}

/*******
Initialize the screen into the proper displaymode.  I'm not sure if
this will be used for any displaymode other than ANSI, so I may be removing
that first "if" eventually.
*******/
void initialize_screen(void){
	if(main_displaymode == ansi){
		initscr();
		getmaxyx(stdscr, yscreensize, xscreensize);

		if(xscreensize < 80 || yscreensize < 25){
			endwin();
			cout << "Error: Screen must be a minimum size of 80 X 25.\n";
			exit(1);
		}

		noecho();
		/*raw();*/

		/********* initialize colours *********/
		start_color();
		init_pair(1, COLOR_BLACK, COLOR_WHITE);
		init_pair(2, COLOR_RED, COLOR_WHITE);
		init_pair(3, COLOR_WHITE, COLOR_BLACK);
		init_pair(4, COLOR_BLACK, COLOR_YELLOW);
		init_pair(CARDBACK_COLOUR, COLOR_GREEN, COLOR_BLUE);
		init_pair(PLAYER_TEXT_COLOUR, COLOR_GREEN, COLOR_BLACK);
		init_pair(PLAYER_TEXT_COLOUR + 1, COLOR_YELLOW, COLOR_BLACK);
		init_pair(MESSAGE_TEXT_COLOUR, COLOR_CYAN, COLOR_BLACK);
		numcolourpairs = 8;
		/*bkgd(COLOR_PAIR(3));*/

		curs_set(0);
		clear();
	}
}

/****
Waits for the user to strike a key and returns it's value.
****/
char wait_for_keypress(void){
	char returnval = (char)0;

	if(main_displaymode == ansi) returnval = wgetch(messagewindow);

	return returnval;
}

/****
This will draw the board (pointed to by *board) on the screen at
location {x, y}, (or {y, x} depending on the method of display).
This will have different "else if"'s for each display mode.
****/
void draw_board(boardtype *board, int starty, int startx){
	int x, y;

	if(main_displaymode == ansi){
		for(y = 0; y < board->ysize; y++){
			move(starty + y, startx);
			for(x = 0; x < board->xsize; x++){
				attron(COLOR_PAIR(board->colour[y][x]));
				addch(board->image[y][x]);
				attroff(COLOR_PAIR(board->colour[y][x]));
			}
		}
	}
}
/*****
Draw all the pegs on the board in their appropriate locations.  This will
eventually have a different if{} structure for each displaymode.
*****/
void draw_pegs(void){
	int n, x, y;

	if(main_displaymode == ansi){
		for(n = 0; n < numplayers; n++){
			if(player[n].score > 0){
				attron(COLOR_PAIR(board.pegcolour[n]) | A_BOLD);
				if(player[n].score < board.numholes + 1){

					x = board.hole[n][player[n].score - 1].x;
					y = board.hole[n][player[n].score - 1].y;
					move(y + boardy, x + boardx);
					addch('!');

					if(player[n].priorscore > 0){
						x = board.hole[n][player[n].priorscore - 1].x;
						y = board.hole[n][player[n].priorscore - 1].y;
						move(y + boardy, x + boardx);
						addch('!');
					}
				}else{
					x = board.endhole.x;
					y = board.endhole.y;
					move(y + boardy, x + boardx);
					addch('!');

					x = board.hole[n][player[n].priorscore - 1].x;
					y = board.hole[n][player[n].priorscore - 1].y;
					move(y + boardy, x + boardx);
					addch('!');
				}
				attroff(COLOR_PAIR(board.pegcolour[n]) | A_BOLD);
			}
			if(main_displaymode == ansi) wrefresh(messagewindow);
		}
	}
}

/***
Increase "player[playernum]"'s score by "amount".
This function should eventually include the option of animating the pegs
being moved forward (one hole at a time), as the score is increased.
****/
void increment_score(int playernum, int amount){
	/***** increment the score and assign it's old value to "priorscore" ******/
	player[playernum].priorscore = player[playernum].score;
	player[playernum].score += amount;
}

/****
Take all of the cards which are dealt out and put them back in the deck.
*****/
void restock_deck(void){
	int n;
	for(n = 0; n < numplayers; n++){
		deck += player[n].hand;
		player[n].hand.erase();
	}
	deck += crib;
	crib.erase();
	deck += cutcard;
	cutcard.erase();
}

/*****
Indicate the beginning of a new round and show that the current round
number is "numrounds".  The method of display will vary depending on 
what kind of user interface is in use at the moment.
*****/
void display_round_start(int numrounds){
	if(main_displaymode == ansi){
		wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD);
		wprintw(messagewindow, "\n=== Round %i =============\n", numrounds);
		wprintw(messagewindow, "%s deals.\n", player[current_dealer].name);
		wrefresh(messagewindow);
		wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD);
	}
}

/******
Get the names of the various players.
******/
void get_player_names(void){
	int n, m;
	if(playstyle == client){
		if(main_displaymode == ansi){
			clear();
			move(0, 0);
			curs_set(1);
			printw("Hi there, what is your name? ", n + 1);
			echo();
			scanw("%s", localplayername);
			noecho();
			curs_set(0);
			clear();
		}
	}else{
		for(n = 0; n < numplayers; n++){
			if(player[n].typeofplayer != console){
				do{
					player[n].getname();
					for(m = 0; m < n && strcmp(player[n].name, player[m].name); m++);
				}while(m < n);
			}else{
				if(main_displaymode == ansi){
					clear();
					move(0, 0);
					curs_set(1);
					printw("Hi there, what is your name? ", n + 1);
					echo();
					scanw("%s", player[n].name);
					noecho();
					curs_set(0);
					clear();
				}
			}
		}
	}
}

/*****
Show a celebratory message indicating who has won the game.  The way in 
which it's displayed will vary greatly depending on what kind of user interface
is in use at the moment.
*****/
void show_winner(displaymodetype displaymode){
	if(displaymode == ansi){
		wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD | A_BLINK);
		wprintw(messagewindow, "%s wins!\n",  player[winning_player].name);
		wrefresh(messagewindow);
		wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD | A_BLINK);
	}

}

/*****
Draw the back side of a card.  The way in which it's drawn will vary
greatly depending on what kind of user interface is in use at the moment.
*****/
void draw_cardback(int y, int x){
	if(main_displaymode == ansi){
		move(y, x);
		attron(COLOR_PAIR(CARDBACK_COLOUR));
		printw("\\|/");
		move(y + 1, x);
		printw("-X-");
		move(y + 2, x);
		printw("/|\\");
		attroff(COLOR_PAIR(CARDBACK_COLOUR));	
	}
}

/******
Deal with command-line arguments.
A return value of 0 indicates that all is good and we're ready to go.
A return value of -1 indicates that all is good, but the game should not be played.
A return value of 1 indicates an error in either understanding or following arguments.
******/
int handle_arguments(int argc, char *argv[]){
	int n, returnval = 0, showhelp = 0;
	FILE *testfile;
		  
	if(argc > 1){
		for(n = 1; n < argc; n++){
			if(!strcmp(argv[n], "--version")){
				printf("Netcrib 0.1\n");
				printf("Copyright (C) 2002 Jacob A Ewing\n");
				printf("Netcrib comes with NO WARRANTY,\n");
				printf("to the extent permitted by law.\n");
				printf("You may redistribute copies of Netcrib\n");
				printf("under the terms of the GNU General Public License.\n");
				printf("For more information about these matters,\n");
				printf("visit www.gnu.org\n");

				returnval = -1;
				break;
			}else if(!strcmp(argv[n], "--help")){
				showhelp = 1;
				returnval = -1;
				break;
			}else if(!strcmp(argv[n], "-d")){
				if(argc > n + 1){
					n++;
					if(!strcmp(argv[n], "none")) main_displaymode = none;
					if(!strcmp(argv[n], "ansi")) main_displaymode = ansi;
				}else{
					returnval = 1;
					break;
				}
			}else if(!strcmp(argv[n], "-h") || !strcmp(argv[n], "--host")){
				playstyle = host;
			}else if(!strcmp(argv[n], "-c") || !strcmp(argv[n], "--client")){
				playstyle = client;
			}else if(!strcmp(argv[n], "-l") || !strcmp(argv[n], "--local")){
				playstyle = local;
			}else if(!strcmp(argv[n], "-b") || !strcmp(argv[n], "--board")){
				if(argc > n + 1){
					testfile = fopen(argv[n + 1], "rb");
					if(testfile != NULL){
						fclose(testfile);
						free(boardfilename);
						boardfilename = strdup(argv[n + 1]);
						n++;
					}else{
						cout << "Cannot find specified board file.\n";
						returnval = 2;
					}
				}else{
					returnval = 1;
					break;
				}
			}else{
				returnval = 1;
			}
		}
	}


	if(returnval == 1 || showhelp){
		cout << "Command Syntax:\n";
		cout << "Netcrib [(-c | --client) | (-h | --host) | (-l | --local)]\n";
		cout << "        [-d (display mode)]\n";
		cout << "        [(-b | --board) filename]\n";
		cout << "\nOPTIONS:\n";
		cout << "\t-c or --client\n\t\tAllows you to connect to another machine\n";
		cout << "\t\twhich is serving a game.\n";
		cout << "\t-h or --host\n\t\tStarts a game in server mode, and will\n";
		cout << "\t\twait for enough players to connect.\n";
		cout << "\t-l or --local (default)\n\t\tStarts a local human vs. computer game.\n";
		cout << "\t-d (display mode)\n\t\tSelects which type of interface you want to use,\n\t\tand can be one of the following:\n";

		cout << "\t\t\tnone: Will not display anything.\n";
		cout << "\t\t\tansi: Displays the entire game in ansi highlighted text.\n";
		cout << "\t-b (filename) or --board (filename)\n\t\tLoad the board information from the file (filename).\n";
	}

	return returnval;
}

/*******
Load the cribbage board info from the file whose name is pointed to by
*filename, and initialize some global variables with the values loaded.
Returns 0 upon success and 1 if the file can not be opened.
*******/
int load_board(char *filename){
	int returnval = 0, n, x, y;

	if(board.load(filename)){
		returnval = 1;
	}else{
		numplayers = board.numplayers;			/* The number of players in the game */
		current_dealer = rand() % numplayers;		/* The index number of the player who's dealing */
		maximum_score = board.numholes + 1;		/* The number of points required to win the game */
		handsize = board.handsize;			/* The size of each player's hand after putting cards in the crib */
		add_to_crib_size = board.add_to_crib_size;	/* The number of cards each player needs to throw in the crib */
		starting_handsize = board.starting_handsize;	/* The number of cards initially dealt to each player */
		starting_cribsize = board.starting_cribsize;	/* The number of cards initially dealt into the crib */

		baseboardcolour = numcolourpairs + 1;
		for(n = 0; n < board.numcolourpairs; n++){
			numcolourpairs++;
			init_pair(numcolourpairs, board.colourpair[n][0], board.colourpair[n][1]);
		}

		for(y = 0; y < board.ysize; y++){
			for(x = 0; x < board.xsize; x++){
				board.colour[y][x] += baseboardcolour;
			}
		}

		for(n = 0; n < board.numplayers; n++){
			board.pegcolour[n] += baseboardcolour;
		}

		boardy = 1;
		boardx = xscreensize - board.xsize - 1;
	}
	return returnval;
}

/****
Set up different output areas.  This will differ greatly depending on what
kind of user interface is used.
****/
void initialize_windows(void){
	int height, width;
	if(main_displaymode == ansi){
		messagey = boardy + board.ysize;
		messagex = boardx;

		height = yscreensize - messagey - 2;
		width = board.xsize;

		messagewindow = newwin(height, width, messagey, messagex);
		scrollok(messagewindow, 1);	
	}
}

/******
Update the screen.  This will depend on what method of display the player
is using.
******/
void redraw_screen(void){
	if(main_displaymode == ansi){
		draw_board(&board, boardy, boardx);
		draw_pegs();
		refresh();
		wrefresh(messagewindow);
	}
}

/*****
Goes through each of the players' hands and counts their point value along
with the cut card.  It does this in the correct order of play (ie. Start with
the player to the "left" of the dealer, counting the dealer's hand and crib
last.  If any players reach the end of the board the game ends.  Visual
output depends on the value of "displaymode".
*****/
void peg_hands(displaymodetype displaymode){
	int n, m;
	for(n = 0; n < numplayers; n++){

		m = (n + current_dealer + 1) % numplayers;
		if(displaymode == ansi){
			wattron(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + m) | A_BOLD);
			wprintw(messagewindow, "\n%s pegs:\n", player[m].name);
			wattroff(messagewindow, A_BOLD);
		}
		increment_score(m, count_points(player[m].hand, HAND, displaymode));

		redraw_screen();
		if(displaymode == ansi){
			wattroff(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + m));
			wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
			wprintw(messagewindow, "Press any key to continue.\n\n");
			wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
			wrefresh(messagewindow);
			wait_for_keypress();
		}

		if(player[m].score >= maximum_score){
			game_over = 1;
			winning_player = m;
			break;
		}
	}

	if(!game_over){
		if(displaymode == ansi){
			wattron(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + current_dealer) | A_BOLD);
			wprintw(messagewindow, "And in the crib:\n");
			wattroff(messagewindow, A_BOLD);
			if(current_dealer == 0){
				crib.display(yscreensize - 5, 0);
			}else{
				box(1, 0, (numplayers - 1) << 2, boardx - 1, ' ');
				crib.display(2, 0);
			}
			redraw_screen();
		}

		m = count_points(crib, CRIB, displaymode);
		if(m){
			increment_score(current_dealer, m);
			redraw_screen();
		}

		if(player[current_dealer].score >= maximum_score){
			game_over = 1;
			winning_player = current_dealer;
		}
	}

	redraw_screen();
}

/*****
count_play_value counts and returns the value of the card played
by player[current_player].  Visual output depends on the value
of displaymode.
*****/
int count_play_value(handtype pile, int current_player, displaymodetype displaymode){
	int n, m, sum, returnval = 0, testsize, runsize;
	handtype temphand;

	/******* check for fifteens or thirty-ones *******/
	sum = pile.sum();
	if(sum == 15){
		returnval += 2;
		if(displaymode == ansi) wprintw(messagewindow, "Fifteen two");
	}else if(sum == 31){ 
		returnval += 2;
		if(displaymode == ansi) wprintw(messagewindow, "Thirty-one for two");
	}else if(pile.numcards == numplayers * handsize){
		returnval++;
		if(displaymode == ansi) wprintw(messagewindow, "Go for one");
	}else{
		if(displaymode == ansi) wprintw(messagewindow, "%i", sum);
	}

	/******* check for pairs, pair royals, etc. ********/
	m = 1;
	for(n = pile.numcards - 1; n > 0 && pile.card[n].value == pile.card[n - 1].value; n--) m++;
	if(m >= 2){
		returnval += (m * (m - 1));

		if(displaymode == ansi){
			if(m == 2)
				wprintw(messagewindow, ", pair for two");
			else if(m == 3)
				wprintw(messagewindow, ", pair royal for %i",  m * (m - 1));
			else
				wprintw(messagewindow, ", double royal for %i",  m * (m - 1));
		}
	}

	/******** check for runs *******/
	if(pile.numcards > 2){
		runsize = 0;
		temphand.erase();

		temphand += pile.card[pile.numcards - 1];
		temphand += pile.card[pile.numcards - 2];

		for(testsize = 3; testsize <= pile.numcards; testsize ++){
			temphand += pile.card[pile.numcards - testsize];
			temphand.sort();
			m = 1;
			for(n = 1; n < testsize; n++){
				if(temphand.card[n].value - temphand.card[n - 1].value == 1){
					m++;
				}else{
					break;
				}
			}
			if(n >= testsize){
				runsize = m;
			}
		}
		if(runsize > 2){
			returnval += runsize;
			if(displaymode == ansi) wprintw(messagewindow, ", Run for %i", runsize);
		}
	}
	if(displaymode == ansi) wprintw(messagewindow, ".\n");
	return returnval;
}

/*****
This is where the hands are actually played out.
*****/
void play_hands(void){
	int x, y, lotsofspace;
	int cardsplayed[numplayers], passedonce[numplayers];
	int total_cards_played = 0, current_numcards = 0, n;
	int current_player, last_player;
	handtype pile, original_hand[numplayers];

	for(n = 0; n < numplayers; n++){
		original_hand[n] = player[n].hand;
		cardsplayed[n] = passedonce[n] = 0;
	}

	pile.erase();
	current_player = (current_dealer + 1) % numplayers;

	while(total_cards_played < numplayers * handsize && !game_over){
		if(main_displaymode == ansi)
			wattron(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + current_player) | A_BOLD);
		if((passedonce[current_player] != 1) && (player[current_player].play_card(&pile))){

			if(main_displaymode == ansi){
				lotsofspace = ((numplayers - 1) << 3) < (yscreensize >> 1) - 1;
				
				x = (cardsplayed[current_player] << 2);
				if(current_player == 0) y = (yscreensize >> 1) + 3;
				else{
					if(!lotsofspace)
						y = 1 + ((current_player - 1) << 2);
					else						
						y = (yscreensize >> 1) - (current_player << 2) - 1;
				}

				cardsplayed[current_player]++;
				pile.card[current_numcards].display(y, x);

				if(lotsofspace){
					if(current_player == 0) y = yscreensize - 3;
					else y = 1 + ((current_player - 1) << 2);
					x = (handsize - cardsplayed[current_player]) << 2;
					box(y, x, 3, 3, ' ');
				}
			}

			current_numcards++;
			total_cards_played ++;

			n = count_play_value(pile, current_player, main_displaymode);
			if(n != 0){
				increment_score(current_player, n);
				redraw_screen();
			}

			if(player[current_player].score >= maximum_score){
				game_over = 1;
				winning_player = current_player;
				redraw_screen();

			}

			last_player = current_player;
		}else{

			if(last_player == current_player){
				pile.erase();
				current_numcards = 0;
				pile.numcards = 0;

				increment_score(current_player, 1);
				redraw_screen();

				if(player[current_player].score >= maximum_score){
					game_over = 1;
					winning_player = current_player;
				}

				if(main_displaymode == ansi){
					wprintw(messagewindow, "Go for one\n");
					wrefresh(messagewindow);
				}

				for(n = 0; n < numplayers; n++){
					passedonce[n] = 0;
				}

			}else if(!passedonce[current_player]){
				if(main_displaymode == ansi){
					wprintw(messagewindow, "Pass\n");
					wrefresh(messagewindow);
				}
				passedonce[current_player] = 1;

			}
		}

		if(main_displaymode == ansi)
			wattroff(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + current_player) | A_BOLD);

		if(pile.sum() == 31){
			pile.erase();
			current_numcards = 0;
			for(n = 0; n < numplayers; n++)
				passedonce[n] = 0;
		}

		current_player = (current_player + 1) % numplayers;
		if(main_displaymode == ansi){
			wrefresh(messagewindow);
			refresh();
		}

	}

	if(pile.numcards && !game_over){
		increment_score(last_player, 1);

		if(main_displaymode == ansi){
			wattron(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + last_player) | A_BOLD);
			wprintw(messagewindow, "One for last card.\n");
			wattroff(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + last_player) | A_BOLD);
		}
	}

	if(main_displaymode == ansi){
		/**** erase the last card showing in the human player's hand and the text below ***/
		box(yscreensize - 6, 0, 5, 20, ' ');
		
		/*** update the board image ***/
		draw_board(&board, boardy, boardx);
		draw_pegs();
		refresh();

		wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
		wprintw(messagewindow, "Press any key to continue.\n\n");
		wrefresh(messagewindow);
		wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));

		wait_for_keypress();

	}
	for(n = 0; n < numplayers; n++){
		player[n].hand = original_hand[n];
	}

	redraw_screen();

}

/*****
This deals "numtodeal" cards from the deck "d" to the hand which is
held by the player "p".
*****/
void dealfromdeck(decktype *d, playertype *p, int numtodeal){
	int n;
	for(n = 0; n < numtodeal; n++){
		if(p->hand.numcards < MAXHANDSIZE && d->numcards > 0){
			d->numcards--;
			p->hand.card[p->hand.numcards] = d->card[d->numcards];
			d->card[d->numcards].erase();
			p->hand.numcards++;
		}
	}
}

/****
Deal cards from the deck into each of the player's hands.  You'll notice
that this follows the same order in which a human would deal the cards.
If the crib gets any cards dealt into it, that is done here as well.
****/
void deal_round(void){
	int n, m, l;

	for(n = 0; n < starting_handsize; n++){
		for(m = 0; m < numplayers; m++){
			l = (m + current_dealer) % numplayers;
			dealfromdeck(&deck, &player[l], 1);
		}
	}

	for(n = 0; n < starting_cribsize; n++){
		deck.numcards --;
		crib += deck.card[deck.numcards];
		deck.card[deck.numcards].erase();
	}
}

/******
Cut the deck and check for nobs.
******/
void cut_deck(void){
	int range, cutnum, minimum, x, y;
	minimum = 4;
	range = deck.numcards - (minimum << 1);
	cutnum = minimum + rand() % range;
	cutcard = deck.draw(minimum + rand() % range);
	
	if(main_displaymode == ansi){
		x = 4;
		y = (yscreensize >> 1) - 2;
		mvprintw(y, x, "Cut:");
		x = 5;
		y = (yscreensize >> 1) - 1;
		cutcard.display(y, x);
	}

	if(cutcard.value == JACK){
		increment_score(current_dealer, 1);
		if(player[current_dealer].score >= maximum_score){
			game_over = 1;
			winning_player = current_dealer;
		}
		if(main_displaymode == ansi){
			x = 0;
			y = (yscreensize >> 1) + 3;
			wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD);
			wprintw(messagewindow, "Dealer gets one for Nobs.\n");
			wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD);
			wrefresh(messagewindow);
			draw_pegs();
		}
	}
	redraw_screen();
}

/*****
Set the current time as the random number key.
*****/
void randomize(void){
	srand(time(0));
}

/****
Have each of the players decide which cards to put in the crib.
****/
void divide_hands(void){
	int n, m, x;
	
	/******* Draw the backs of the cards for all opponents ******/
	if(main_displaymode == ansi){
		for(n = 0; n < starting_handsize; n++){
					for(m = 0; m < numplayers - 1; m++){
				draw_cardback(1 + (m << 2), n << 2);
			}
		}
	}
	
	for(n = 0; n < numplayers; n++){
		player[n].divide_hand();
	}

	/****** Erase the backs of the number of cards each opponent put in the crib **********/
	if(main_displaymode == ansi){
		x = (starting_handsize - add_to_crib_size) * 4;
		box(1, x, (numplayers - 1) << 2, add_to_crib_size << 2, ' ');
	}
}

/*****
Counts points provided by fifteens in the handtype object "hand" and returns
that value.  playmode describes what part of the game this is being counted
for (eg. points in the crib, pegging, etc.).  displaymode determines what
kind of output goes on the screen.
*****/
int count_fifteens(handtype hand, playmodetype playmode, displaymodetype displaymode){
	int score, numcombos, tally, n, m;
	handtype temphand;

	temphand = hand;
	if(playmode == HAND || playmode == CRIB) temphand += cutcard;
	temphand.sort();

	/************ looking for fifteens. ***********/
	numcombos = 1 << temphand.numcards;   /* This is the total number of combinations of the cards */
	score = 0;

	for(n = 1; n < numcombos; n++){
		tally = 0;
		for(m = 0; m < temphand.numcards; m++){
			if((1 << m) & n){
				tally += temphand.card[m].pointvalue;
			}
			
		}

		if(tally == 15){

			if(score != 0 && displaymode == ansi)
				wprintw(messagewindow, ", ");
			score += 2;
			if(displaymode == ansi) 
				wprintw(messagewindow, "Fifteen %i", score);
		}
	}
	if(displaymode == ansi && score != 0){
		wprintw(messagewindow, "\n");
		wrefresh(messagewindow);
	}

	return score;
}

/*****
Counts points provided by pairs and pair royals in the handtype object "hand"
and returns that value.  playmode describes what part of the game this is
being counted for (eg. points in the crib, pegging, etc.).  displaymode
determines what kind of output goes on the screen.
*****/
int count_pairs(handtype hand, playmodetype playmode, displaymodetype displaymode){
	int n, m, alreadycounted, tally, score = 0, numpairs = 0, spoken = 0;
	handtype temphand;


	temphand = hand;
	if(playmode == HAND || playmode == CRIB) temphand += cutcard;


	/********** looking for pairs and pair royals **********/
	for(n = 0; n < temphand.numcards; n++){
		/*** determine if the current card has already been counted in a pair combo ***/
		alreadycounted = 0;
		for(m = 0; m < n; m++){
			if(temphand.card[m].value == temphand.card[n].value)
				alreadycounted = 1;
		}

		if(!alreadycounted){
			tally = 1;
			for(m = n + 1; m < temphand.numcards; m++){
				if(temphand.card[n].value == temphand.card[m].value)
					tally++;
			}

			if(displaymode == ansi && tally > 1){
				switch(tally){
				    case 2: numpairs++; break;
				    case 3:
					if(!spoken) wprintw(messagewindow, ", ");
					wprintw(messagewindow, "Pair royal for six");
					spoken = 1;
					break;
				    case 4: 
					if(!spoken) wprintw(messagewindow, ", ");
					wprintw(messagewindow, "Double royal for twelve");
					spoken = 1;
					break;
				}
			}

			/** no if here, since tally - 1 == 0 if there are no pairs **/
			score += tally * (tally - 1);
		}
	}

	if(displaymode == ansi && score != 0){
		switch(numpairs){
		    case 0: break;
		    case 1:
			if(score > (numpairs << 1)) 
				wprintw(messagewindow, ", and a pair for two.");
			else
				wprintw(messagewindow, "Pair for two");
			break;
		    case 2:
			if(score > (numpairs << 1)) wprintw(messagewindow, ", ");
			if(rand() % 2)
				wprintw(messagewindow, "Pair of pairs for four");
			else
				wprintw(messagewindow, "Two pairs for four");
			break;
		    default:
			if(score > (numpairs << 1)) wprintw(messagewindow, ", ");
			wprintw(messagewindow, "%i pairs for %i", numpairs, numpairs << 1); break;


		}
		wprintw(messagewindow, "\n");
		wrefresh(messagewindow);
	}
	
	return score;
}

/*****
Counts points provided by runs in the handtype object "hand" and returns
that value.  playmode describes what part of the game this is being counted
for (eg. points in the crib, pegging, etc.).  displaymode determines what
kind of output goes on the screen.
*****/
int count_runs(handtype hand, playmodetype playmode, displaymodetype displaymode){
	int n, runlength, multiplicand, newfactor, score, increment;
	handtype temphand;

	score = 0;
	runlength = multiplicand = newfactor = 1;

	/* is is absolutely necessary that the hand be sorted,
	 * otherwize, this algorithm won't work.
	 */	
	temphand = hand;
	if(playmode == HAND || playmode == CRIB) temphand += cutcard;
	temphand.sort();

	for(n = 1; n < temphand.numcards; n++){
		if(temphand.card[n - 1].value == temphand.card[n].value){
			/*** if the current card is the same number as the previous***/
			newfactor ++;
		}else{
			if(newfactor > 1){
				multiplicand *= newfactor;
				newfactor = 1;
			}

			if(temphand.card[n - 1].value == temphand.card[n].value - 1){
				/*** if the curent card is one up from the previous **/
				runlength++;
			}else{
				if(runlength > 2){
					increment = runlength;
					increment *= multiplicand;
					score += increment;

					if(displaymode == ansi){
						switch(multiplicand){
						    case 2: wprintw(messagewindow, "Double run "); break;
						    case 3: wprintw(messagewindow, "Triple run "); break;
						    case 4: wprintw(messagewindow, "Quadruple run "); break;
						    default: wprintw(messagewindow, "Run ");
						}
						wprintw(messagewindow, "for %i", increment);
					}
				}
				multiplicand = newfactor = runlength = 1;
			}
		}
	}

	/***
	This repetition of the "if" needs to be done in case the loop
	ends with a run at the end of the hand.
	***/
	if(runlength > 2){
		if(newfactor > 1) multiplicand *= newfactor;
		increment = runlength;
		increment *= multiplicand;
		score += increment;

		if(displaymode == ansi){
			switch(multiplicand){
			    case 2: wprintw(messagewindow, "Double run "); break;
			    case 3: wprintw(messagewindow, "Triple run "); break;
			    case 4: wprintw(messagewindow, "Quadruple run "); break;
			    default: wprintw(messagewindow, "Run ");
			}
			wprintw(messagewindow, "for %i", increment);
		}
	}

	if(score > 0 && displaymode == ansi){
		wprintw(messagewindow, "\n");
		wrefresh(messagewindow);
	}

	return score;
}

/*****
Counts points provided by a flush in the handtype object "hand" and returns
that value.  playmode describes what part of the game this is being counted
for (eg. points in the crib, pegging, etc.).  displaymode determines what
kind of output goes on the screen.
*****/
int count_flush(handtype hand, playmodetype playmode, displaymodetype displaymode){
	int score = 0, n, isaflush;

	isaflush = (playmode != PEG);
	for(n = 1; n < hand.numcards && isaflush; n++){
		if(hand.card[n].suit != hand.card[n - 1].suit)
			isaflush = 0;
	}

	if(isaflush){
		if(playmode == CRIB && cutcard.suit != hand.card[0].suit) isaflush = 0;
	}

	if(isaflush){
		score = hand.numcards;
		if(cutcard.suit == hand.card[0].suit) score++;
	}

	if(score != 0  && displaymode == ansi){
		wprintw(messagewindow, "%i card flush\n", score);
		wrefresh(messagewindow);
	}
	return score;
}

/*****
Counts points provided by "nobs" in the handtype object "hand" and returns
that value.  playmode describes what part of the game this is being counted
for (eg. points in the crib, pegging, etc.).  displaymode determines what
kind of output goes on the screen.
*****/
int count_nobs(handtype hand, playmodetype playmode, displaymodetype displaymode){
	int returnval = 0, n;

	if(cutcard.value != JACK && playmode != PEG){
		for(n = 0; n < hand.numcards; n++){
			if(hand.card[n].value == JACK
			   && hand.card[n].suit == cutcard.suit)
				returnval = 1;
		}
	}

	if(displaymode == ansi && returnval != 0){
		wprintw(messagewindow, "1 for his nobs\n");
		wrefresh(messagewindow);
	}

	return returnval;
}

/******
Counts and returns the number of points in the handtype "hand".
playmode determines the manner in which the points are counted.
For example, if playmode == CRIB, then a flush will only be counted
if the card that was cut is the same suit as well.  Similarily, if
playmode == PEG, then flushes and nobs would not be counted at all.
displaymode determines the what kind of output is displayed on the
screen.
******/
int count_points(handtype hand, playmodetype playmode, displaymodetype displaymode){
	/***
	This goes through every possible type of score and checks
	to see if any of them can be found in the hand.  eg.  first
	it looks for anything that adds up to fifteen, then it looks
	for runs, and so on.  The individual checks are performed
	by separate functions.
	***/
	int score;

	score = count_fifteens(hand, playmode, displaymode);
	score += count_runs(hand, playmode, displaymode);
	score += count_pairs(hand, playmode, displaymode);
	score += count_nobs(hand, playmode, displaymode);
	score += count_flush(hand, playmode, displaymode);	

	/*** display the score ***/
	if(displaymode == ansi){
		if(score == 0)
			switch(rand() % 8){
				case 0: wprintw(messagewindow, "Good old \"Nineteen\".\n"); break;
				case 1: wprintw(messagewindow, "Nothing whatsoever.\n"); break;
				case 2: wprintw(messagewindow, "Zero, Zip and Zilch.\n"); break;
				case 3: wprintw(messagewindow, "Absolutely nothing.\n"); break;
				case 4: wprintw(messagewindow, "Diddely-Squat.\n");break;
				case 5: wprintw(messagewindow, "Nothing.\n");break;
				case 6: wprintw(messagewindow, "Nada.\n");break;
				default: wprintw(messagewindow, "Nothing at all.\n");
			}

		else{
			wprintw(messagewindow, "for a total of %i.\n", score);
		}
		wrefresh(messagewindow);
	}

	return score;
}



/*** 
Decide which card to play while pegging hands.
*pile is a pointer to the cards that have already been played.
This returns a 1 or 0 depending on whether or not it can play a card
without going over 31, and if so, it puts that card into the hand "pile".
it does NOT add to the score.  This is done elsewhere.
***/
int playertype::play_card(handtype *pile){
	int returnval = 0, n, bestscore = -1, newscore, bestplaycard = -1, tally, cardpicked;
	handtype testhand;
	int gotone;
	unsigned char keypressed;

	if(pile->numcards < MAXHANDSIZE){

		/*******************************************
		This part of the if handles User I/O if the
		player is a human one.
		*******************************************/
		if(typeofplayer == console){
			if(hand.numcards > 0){

				/** first, display the cards and options for the player to see **/
				box(yscreensize - 6, 0, 4, (starting_handsize << 2), ' ');
				wprintw(messagewindow, "%s: ", name);

				for(n = 0; n < hand.numcards; n++){
					hand.card[n].display(yscreensize - 6, n * 4);
					move(yscreensize - 3, n * 4);
					printw("[%c] ", (char)((int)'a' + n));
				}
				printw("\nHit space to pass");
				refresh();
				wrefresh(messagewindow);

				gotone = 0;
				do{

					keypressed = getch();
					if(keypressed >= 'A' && keypressed <= 'Z') keypressed += (int)'a' - (int)'A';
					if(keypressed >= 'a' && keypressed < (char)((int)'a' + hand.numcards)) gotone = 1;
					if(keypressed == ' ') gotone = 1;

					/******* now check to see if it's a valid option ******/
					if(gotone){
						cardpicked = (int)(keypressed - 'a');
						tally = pile->sum();
						if(keypressed == ' '){
							/**** make sure they actually can't play a card ****/
							for(n = 0; n < hand.numcards && (tally + hand.card[n].pointvalue > 31); n++);
							if(n < hand.numcards) gotone = 0;
						}else{
							/**** make sure the selected card can actually be played ****/
							newscore = tally + hand.card[cardpicked].pointvalue;
							if(newscore > 31) gotone = 0;
						}
					}
				}while(!gotone);

				if(keypressed != ' '){
					returnval = 1;
					pile->card[pile->numcards] = hand.card[cardpicked];
					pile->numcards ++;
					hand.removecard(cardpicked);
					returnval = 1;
				}
			}

		/*******************************************
		This part of the if handles the computer player.
		*******************************************/
		}else{
			if(main_displaymode == ansi){
				wprintw(messagewindow, "%s: ", name);
				wrefresh(messagewindow);
			}
			tally = pile->sum();
			for(n = 0; n < hand.numcards; n++){
				if(tally + hand.card[n].pointvalue <= 31){
					testhand = *pile;
					testhand += hand.card[n];
					newscore = count_play_value(testhand, 0, none);
					if(newscore > bestscore){
						bestplaycard = n;
						bestscore = newscore;
					}
				}
			}

			if(bestscore == 0){
				do{
					n = rand() % hand.numcards;
				}while(tally + hand.card[n].pointvalue > 31);
				bestplaycard = n;
			}

			if(bestplaycard != -1){
				pile->card[pile->numcards] = hand.card[bestplaycard];
				pile->numcards++;
				hand.removecard(bestplaycard);
				returnval = 1;
			}
		}
	}else{
		wprintw(messagewindow, "\n<<<<<< ERROR: Not enough space in the pile >>>>>>\n");
		wprintw(messagewindow, "Press [ESC] to close");
		wrefresh(messagewindow);
		while(getch() != 27);
		exit(4);
	}
	return returnval;
}

/***
Let the human player decide which cards to pull out of the hand
and throuw in the crib.
***/
void playertype::divide_by_human(void){
	unsigned char keypressed;
	int n, m, done = 0;
	displaymodetype displaymode = main_displaymode;
	int numcardspicked = 0, card_picked[MAXHANDSIZE];

	if(add_to_crib_size){
		hand.sort();

		/** first, display the cards and options for the player to see **/
		if(displaymode == ansi){

			for(n = 0; n < hand.numcards; n++){
				hand.card[n].display(yscreensize - 6, n * 4);
				move(yscreensize - 3, n * 4);
				printw("[%c] ", (char)((int)'a' + n));
			}
			refresh();

			wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
			if(player[current_dealer].myindexnumber != myindexnumber)
				wprintw(messagewindow, "Select which cards to put into %s's crib.\n", player[current_dealer].name);
			else
				wprintw(messagewindow, "Select which cards to put into your crib.\n");

			wprintw(messagewindow, "(press the space bar when ready)\n");
			wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
			wrefresh(messagewindow);

		}

		/*** now let them pick out the cards they pass to the crib ***/
		numcardspicked = 0;
		do{
			keypressed = getch();

			if(keypressed == (char)27){
				endwin();
				exit(1);
			}

			if(keypressed == ' '){
				if(numcardspicked != add_to_crib_size){
					wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD);
					if(numcardspicked < add_to_crib_size)
						wprintw(messagewindow, "You need to put %i cards in the crib.\n", add_to_crib_size);
					else
						wprintw(messagewindow, "You can only put %i cards in the crib.\n", add_to_crib_size);
					wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD);
					wrefresh(messagewindow);
				}else{
					done = 1;
				}
			}

			for(n = 0; n < hand.numcards && (keypressed != (unsigned char)((int)'a' + n)); n++);
			if(n < hand.numcards){
				for(m = 0; m < numcardspicked; m++)
					if(card_picked[m] == n) break;

				if(m >= numcardspicked){
					card_picked[numcardspicked] = n;
					numcardspicked ++;

					attron(COLOR_PAIR(4));

					move(yscreensize - 3, 4 * n);
					printw("[%c]", 'A' + keypressed - 'a');
					attroff(COLOR_PAIR(4));

					hand.card[n].display(yscreensize - 6, n * 4);

				}else{
					move(yscreensize - 3, 4 * n);
					printw("[%c]", keypressed);

					numcardspicked--;
					for(; m < numcardspicked; m++){
						card_picked[m] = card_picked[m + 1];
					}					
				}
				move(0, 0);
			}
		}while(!done);

		/*** cards have been selected, now throw them in the crib ***/
		for(n = 0; n < numcardspicked; n++){
			hand.passcard(&crib, card_picked[n]);
			/** IMPORTANT
			this additional loop needs to be here, because
			the cards get rearranged in the hand when one is removed;
			**/

			for(m = n + 1; m < numcardspicked; m++)
				if(card_picked[m] > card_picked[n]) card_picked[m]--;
		}
	}
}

/**
Lets the computer decide which card to pull out of it's hand and throuw in the crib.
**/
void playertype::divide_by_computer(void){
	int combo, max, n, numbits, bestcombo, bestscore = -1, score;
	handtype testhand;

	max = 1 << hand.numcards;

	for(combo = 0; combo < max; combo++){
		numbits = 0;
		for(n = 0; n < hand.numcards; n++){
			if(combo & (1 << n)) numbits++;
		}
		if(numbits == add_to_crib_size){
			/***Here is where we check the value of the hand ***/
			testhand.erase();

			/** generate the new hand **/
			for(n = 0; n < hand.numcards; n++){
				if(!(combo & (1 << n))){
					testhand += hand.card[n];
				}
			}

			/** get the value of it **/
			score = count_points(testhand, NOCUT, none);
			if(score > bestscore ){
				bestscore = score;
				bestcombo = combo;
			}
		}
	}

	/*** 
	very important:  This loop MUST go in this order because when passcard()
	is called, every card with a higher index than the one pulled is
	shuffled down by one to fill in the empty space. If they are grabbed
	in ascending order, everything after the first one will be wrong.
	***/
	for(n = hand.numcards - 1; n >= 0; n--)
		if((1 << n) & bestcombo)
			hand.passcard(&crib, n);
}

/******
Prompts the user with the text message that *text points to.
It waits for the user to enter one of two characters, those
being defined by option1 and option2.  If case_sensitive is
set to zero, then the function will accept both upper and lower
case versions of the options if they are letters.
It will return the character value of whichever option is selected.
Note that if case_sensitive is set to 0, then the return value will always be in lower-case.
*******/
char binaryprompt(char *text, char option1, char option2, int case_sensitive){
	char c;

	printw("%s", text);
	refresh();
	if(case_sensitive){
		do{
			c = getch();
		}while(c != option1 && c != option2);
	}else{
		option1 = tolower(option1);
		option2 = tolower(option2);
		do{
			c = (char)tolower(getch());
		}while(c != option1 && c != option2);
	}

	return c;
}

/*****
Draws a filled rectangle in ANSI characters.
(y1, x1) is the the top left corner of the rectangle.
ysize and xsize are the dimensions of the rectangle.
value is the character that the rectangle is made with
******/
void box(int y1, int x1, int ysize, int xsize, char value){
	int y, y2;
	char *string;

	y2 = y1 + ysize - 1;
	string = new char[xsize + 1];
	memset((void *)string, value, xsize);
	string[xsize] = (char)NULL;
	
	for(y = y1; y <= y2; y++){
		mvprintw(y, x1, string);
	}

	delete string;
}
