/*****************************************************************************
	When done, this will hopefully be a nice network game with which
	you can play cribbage against other people or against the computer
	itself.  My starting goal is a simple console game where you can
	play cribbage against the computer.  

	The basic console game is finished now.  I'm now going to attempt
	to make it into a network game.  After that I'll be adding a nice GUI
*****************************************************************************/
#include <iostream.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <ncurses.h>

/**
an odd number here:  MAXHANDSIZE is set to the largest possible number of 
cards that can be played before hitting 31 when pegging.  This is because
this program handles pegging by duplicating them in another hand named
"pile" and checking for points in that.
**/
#define ENTER_KEY (char)10
#define decksize 52
#define MAXHANDSIZE 13
#define maxplayers 4
#define MAXNAMELENGTH 16
#define NUMCARDBACKS 2
#define NUMCARDFACES 52

#define JACK 11
#define QUEEN 12
#define KING 13
#define ACE 1

#define CARDBACK_COLOUR 5
#define PLAYER_TEXT_COLOUR 6
#define MESSAGE_TEXT_COLOUR 8

/****************** enumerations *******************/
enum playmodetype{ HAND, CRIB, PEG, NOCUT };

/***** important note ******
The order in which the suits are listed in "suittype" is important.
This is because the images are loaded by a function that generates
them in that order.
***************************/
enum suittype{ Spades, Hearts, Clubs, Diamonds, nil };
enum displaymodetype { none, ansi };

/****************** classes *******************/
class holetype{
    public:
	int x, y;

	holetype(){
		x = y = 0;
	}

	void set(int newy, int newx){
		x = newx;
		y = newy;
	}
};


class boardtype{
    public:
        int xsize, ysize, numholes, numplayers, numcolourpairs, *pegcolour;
	char **image, **colour, **colourpair;
	holetype **hole, endhole;

	boardtype(){
		xsize = ysize = numholes = numplayers = numcolourpairs = 0;
		image = colour = colourpair = (char **)NULL;
		hole = NULL;
		pegcolour = (int *)NULL;
	}

	boardtype(int newysize, int newxsize, int length, int newnumplayers, int newnumcolourpairs){
		initialize(newysize, newxsize, length, newnumplayers, newnumcolourpairs);
	}

	void initialize(int newysize, int newxsize, int length, int newnumplayers, int newnumcolourpairs){

		int n;
		ysize = newysize;
		xsize = newxsize;

		pegcolour = new int[newnumplayers];
		image = new char*[newysize];
		colour = new char*[newysize];
		colourpair = new char*[newnumcolourpairs];

		numcolourpairs = newnumcolourpairs;
		for(n = 0; n < numcolourpairs; n++){
			colourpair[n] = new char[2];
		}

		for(n = 0; n < newysize; n++){
			image[n] = new char[newxsize + 1];
			colour[n] = new char[newxsize + 1];
			memset(image[n], ' ', (newxsize + 1) * sizeof(unsigned char));
			memset(colour[n], 0, (newxsize + 1) * sizeof(char));
		}

		numholes = length;
		numplayers = newnumplayers;

		hole = new holetype*[numplayers];
		for(n = 0; n < numplayers; n++){
			hole[n] = new holetype[numholes];
		}

	}

	~boardtype(){
		delete hole;
		delete image;
		delete colour;
		delete colourpair;
		delete pegcolour;
	}

	int save(char *filename){
		FILE *fout;
		int n, m;

		fout = fopen(filename, "wb");
		if(fout != NULL){
			fwrite((void *)&xsize, sizeof(int), 1, fout);
			fwrite((void *)&ysize, sizeof(int), 1, fout);
			fwrite((void *)&numholes, sizeof(int), 1, fout);
			fwrite((void *)&numplayers, sizeof(int), 1, fout);
			fwrite((void *)&numcolourpairs, sizeof(int), 1, fout);

			for(n = 0; n < ysize; n++)
				fwrite((void *)image[n], sizeof(char), xsize + 1, fout);
			for(n = 0; n < ysize; n++)
				fwrite((void *)colour[n], sizeof(char), xsize + 1, fout);
			for(n = 0; n < numcolourpairs; n++){
				fwrite((void *)colourpair[n], sizeof(char), 2, fout);
			}

			for(m = 0; m < numplayers; m++){
				for(n = 0; n < numholes; n++){
					fputc((int)hole[m][n].x, fout);
					fputc((int)hole[m][n].y, fout);
				}
				fputc((int)pegcolour[m], fout);
			}
			fputc((int)endhole.x, fout);
			fputc((int)endhole.y, fout);

			fclose(fout);

		}else{
			printf("Error opening save file.\n");
			return 1;
		}
		return 0;
	}

	int load(char *filename){
		FILE *fin;
		int n, m;
		int xs, ys, nh, np, nc;

		fin = fopen(filename, "rb");
		if(fin != NULL){
			fread((void *)&xs, sizeof(int), 1, fin);
			fread((void *)&ys, sizeof(int), 1, fin);
			fread((void *)&nh, sizeof(int), 1, fin);
			fread((void *)&np, sizeof(int), 1, fin);
			fread((void *)&nc, sizeof(int), 1, fin);

			initialize(ys, xs, nh, np, nc);

			for(n = 0; n < ysize; n++)
				fread((void *)image[n], sizeof(char), xsize + 1, fin);
			for(n = 0; n < ysize; n++)
				fread((void *)colour[n], sizeof(char), xsize + 1, fin);
			for(n = 0; n < numcolourpairs; n++){
				fread((void *)colourpair[n], sizeof(char), 2, fin);
			}

			for(m = 0; m < numplayers; m++){
				for(n = 0; n < numholes; n++){
					hole[m][n].x = fgetc(fin);
					hole[m][n].y = fgetc(fin);
				}
				pegcolour[m] = fgetc(fin);
			}
			endhole.x = fgetc(fin);
			endhole.y = fgetc(fin);

			fclose(fin);

		}else{
			printf("Error opening load file.\n");
			return 1;
		}
		return 0;
	}

	void draw(int starty, int startx){
		int x, y;
		for(y = 0; y < ysize; y++){
			move(starty + y, startx);
			for(x = 0; x < xsize; x++){
				attron(COLOR_PAIR(colour[y][x]));
				addch(image[y][x]);
				attroff(COLOR_PAIR(colour[y][x]));
			}
		}
	}
};

class cardtype{
    private:
	void displaysuit(void){
		switch(suit){
			case Diamonds: printw("D"); break;
			case Clubs: printw("C"); break;
			case Hearts: printw("H"); break;
			case Spades: printw("S"); break;
			default: printw("E");
		}
	}

	void displayvalue(void){
		if(value < 11 && value > 1)
			printw("%i", value);
		else{
			switch(value){
				case 1: printw("A"); break;
				case 11: printw("J"); break;
				case 12: printw("Q"); break;
				case 13: printw("K"); break;
				default: printw("E");
			}
		}
	}

    public:
	suittype suit;
	int pointvalue;
	int value;

	cardtype(){ erase(); }

	void erase(void){
		suit = nil;
		value = '0';
		pointvalue = 0;
	}


	void display(displaymodetype displaymode){
		int x, y, colournum;
		switch(displaymode){
		    case ansi:
			getyx(stdscr, y, x);
			if(suit == Hearts || suit == Diamonds)
				colournum = 2;
			else
				colournum = 1;

			move(y, x);
			attron(COLOR_PAIR(colournum));
			displayvalue();
			printw(" ");
			if(value != 10) printw(" ");

			move(y + 1, x);
			printw(" ");
			displaysuit();
			printw(" ");

			move(y + 2, x);
			printw(" ");
			if(value != 10) printw(" ");
			displayvalue();

			attroff(COLOR_PAIR(colournum));
		}
	}

	void display(int y, int x){
		int colournum;
		if(suit == Hearts || suit == Diamonds)
			colournum = 2;
		else
			colournum = 1;

		move(y, x);
		attron(COLOR_PAIR(colournum));
		displayvalue();
		printw(" ");
		if(value != 10) printw(" ");

		move(0, 0);
		move(y + 1, x);
		printw(" ");
		displaysuit();
		printw(" ");

		move(0, 0);
		move(y + 2, x);
		printw(" ");
		if(value != 10) printw(" ");
		displayvalue();

		attroff(COLOR_PAIR(colournum));
	}

	cardtype operator = (cardtype newcard){
		suit = newcard.suit;
		value = newcard.value;
		pointvalue = newcard.pointvalue;
		return newcard;
	}

	int operator != (cardtype testcard){
		int returnval = 0;
		if(suit != testcard.suit) returnval = 1;
		if(value != testcard.value) returnval = 1;
		return returnval;
	}

	int operator == (cardtype testcard){
		int returnval = 1;
		if(suit != testcard.suit) returnval = 0;
		if(value != testcard.value) returnval = 0;
		return returnval;
	}
};

class handtype{
    public:
	cardtype card[MAXHANDSIZE];
	int numcards;

	handtype(){
		/*** start with an empty hand ***/
		numcards = 0;
	}

	void erase(void){
		int n;
		for(n = 0; n < MAXHANDSIZE; n++){
			card[n].erase();
		}
		numcards = 0;
	}

	void display(displaymodetype displaymode){
		int n;
		for(n = 0; n < numcards; n++){
			card[n].display(displaymode);
		}
	}

	void display(int y, int x){
		int n;
		for(n = 0; n < numcards; n++){
			card[n].display(y, x + (n << 2));
		}
	}


	int sum(void){
		int n, returnval = 0;
		for(n = 0; n < numcards; n++){
			returnval += card[n].pointvalue;
		}

		return returnval;
	}

	handtype operator = (handtype newhand){
		int n;
		erase();
		numcards = newhand.numcards;
		for(n = 0; n < numcards; n++){
			card[n] = newhand.card[n];
		}
		return newhand;
	}

	cardtype operator += (cardtype newcard){
		if(numcards < MAXHANDSIZE){
			card[numcards] = newcard;
			numcards++;
		}
		return newcard;
	}

	void sort(void){
		cardtype tempcard;
		int n, m;
		for(n = 1; n < numcards; n++){
			m = n;
			/*** Move the value down to the appropriate spot. ***/
			while(m > 0 && card[m].value < card[m - 1].value){
				/*** switch the values around ***/
				tempcard = card[m];
				card[m] = card[m - 1];
				card[m - 1] = tempcard;
				m--;
			}
		}
	}

	void passcard(handtype *recipient, int cardindex){
		*recipient += card[cardindex];
		removecard(cardindex);
	}

	void removecard(int cardindex){
		int n;
		numcards--;
		for(n = cardindex; n < numcards; n++){
			card[n] = card[n + 1];
		}
		card[n].erase();
	}

};

class playertype{
    private:
	void divide_by_human(void);
	void divide_by_computer(void);

    public:
	int score, priorscore, typeofplayer, number;
	handtype hand;
	char name[MAXNAMELENGTH];

	playertype(){
		score = priorscore = typeofplayer = number = 0;
	}

	void getname(void){
		if(typeofplayer == 1 || 1 /********** this "if" needs to be changed later **************/){
			switch(rand() % 26){
				case 0: strcpy(name, "Alfred"); break;
				case 1: strcpy(name, "Bonnie"); break;
				case 2: strcpy(name, "Charles"); break;
				case 3: strcpy(name, "Dominique"); break;
				case 4: strcpy(name, "Edward"); break;
				case 5: strcpy(name, "Francine"); break;
				case 6: strcpy(name, "Ginny"); break;
				case 7: strcpy(name, "Howard"); break;
				case 8: strcpy(name, "Ivan"); break;
				case 9: strcpy(name, "Josephine"); break;
				case 10: strcpy(name, "Kenny"); break;
				case 11: strcpy(name, "Laurel"); break;
				case 12: strcpy(name, "Michael"); break;
				case 13: strcpy(name, "Nora"); break;
				case 14: strcpy(name, "Oliver"); break;
				case 15: strcpy(name, "Penelope"); break;
				case 16: strcpy(name, "Quincy"); break;
				case 17: strcpy(name, "Rebecca"); break;
				case 18: strcpy(name, "Samson"); break;
				case 19: strcpy(name, "Tasha"); break;
				case 20: strcpy(name, "Uriel"); break;
				case 21: strcpy(name, "Victoria"); break;
				case 22: strcpy(name, "William"); break;
				case 23: strcpy(name, "Xena"); break;
				case 24: strcpy(name, "Yolanda"); break;
				case 25: strcpy(name, "Zachary"); break;
			}
		}
	}

	void showhand(displaymodetype displaymode){
		hand.display(displaymode);
		printw("\n");
	}

	void divide_hand(void){
		if(typeofplayer == 0) divide_by_human();
		else divide_by_computer();
	}

	int play_card(handtype *pile);
};

class decktype{
    public:
	cardtype card[decksize];
	int numcards;

	decktype(){
		int n;

		/*** first, create the cards ***/
		numcards = decksize;
		for(n = 0; n < decksize; n++){
			card[n].suit = (suittype)(n % 4);
			card[n].value = (n >> 2) + 1;
			card[n].pointvalue = card[n].value;
			if(card[n].pointvalue > 10) card[n].pointvalue = 10;
		}

		/*** now shuffle them ***/
		shuffle();
	}

	void shuffle(void){
		int n, m, numruns;
		cardtype swapcard;

		for(numruns = 0; numruns < 6; numruns++){
			for(n = 0; n < decksize; n++){

				if(rand() % 2){
					swapcard = card[n];
					for(m = n; m > 0; m--)
						card[m] = card[m - 1];
					card[0] = swapcard;
				}
			}
		}
	}

	void display(displaymodetype displaymode){
		int n;
		for(n = 0; n < numcards; n++){
			card[n].display(displaymode);
			if(n % 4 == 3) printw("\n");
		}
	}

	void deal(playertype *p, int numtodeal){
		int n;
		for(n = 0; n < numtodeal; n++){
			if(p->hand.numcards < MAXHANDSIZE && numcards > 0){
				numcards--;
				p->hand.card[p->hand.numcards] = card[numcards];
				card[numcards].erase();
				p->hand.numcards++;
			}
		}
	}

	cardtype draw(int cutpos){
		cardtype returnval;
		int n;
		if(cutpos >= 0 && cutpos < numcards){
			returnval = card[cutpos];
			numcards--;
			for(n = cutpos; n < numcards; n++){
				card[n] = card[n + 1];
			}
			card[n].erase();
		}
		return returnval;
	}

	cardtype operator += (cardtype addition){
		if(numcards < decksize){
			card[numcards] = addition;
			numcards ++;
		}
		return addition;
	}

	handtype operator += (handtype addition){
		int n;
		if(numcards + addition.numcards <= decksize){
			for(n = 0; n < addition.numcards; n++){
				card[numcards] = addition.card[n];
				numcards ++;
			}
		}
		return addition;
	}
};


/****************** function declarations *********************/
void randomize(void);
int initialize_table(int argc, char *argv[]);
void deal_round(void);
int count_points(handtype hand, playmodetype playmode, displaymodetype displaymode);
int count_fifteens(handtype hand, playmodetype playmode, displaymodetype displaymode);
int count_runs(handtype hand, playmodetype playmode, displaymodetype displaymode);
int count_pairs(handtype hand, playmodetype playmode, displaymodetype displaymode);
int count_flush(handtype hand, playmodetype playmode, displaymodetype displaymode);
int count_nobs(handtype hand, playmodetype playmode, displaymodetype displaymode);
int count_play_value(handtype pile, int current_player, displaymodetype displaymode);
void divide_hands(void);
void cut_deck(void);
void play_hands(void);
void peg_hands(displaymodetype displaymode);
void show_winner(displaymodetype displaymode);
void get_player_names(void);
void display_round_start(int numrounds);
void restock_deck(void);
void draw_cardback(int y, int x);
void draw_pegs(void);
void box(int y1, int x1, int ysize, int xsize, char value);
int load_board(void);
void initialize_windows(void);
void redraw_screen(void);
void increment_score(int playernum, int amount);
void wait_for_keypress(void);

/****************** global variables ******************/
playertype player[maxplayers];
decktype deck;
cardtype cutcard;
int numplayers, handsize, starting_handsize, starting_cribsize, add_to_crib_size, current_dealer;
int game_over, winning_player, maximum_score;
displaymodetype main_displaymode;
handtype crib, stack;
int current_playmode = 0, xscreensize, yscreensize;
boardtype board;
int numcolourpairs, baseboardcolour, boardx, boardy, messagex, messagey;
WINDOW *messagewindow;

/****************** function definitions *********************/
int main(int argc, char *argv[]){
	int tally[2];

	int n, returnval, numrounds, play_again;
	displaymodetype displaymode;
	char c;

	randomize();

	returnval = initialize_table(argc, argv);
	if(returnval) exit(returnval);

	/*** set to one human vs. (numplayers - 1) bots
	Give each of the players a number as well.  This is used
	to tell within the objects themselves just who they are **/

	for(n = 0; n < numplayers; n++){
		player[n].typeofplayer = 1;
		player[n].number = n;
	}
	player[0].typeofplayer = 2;
	tally[0] = tally[1] = 0;

	displaymode = main_displaymode;

	/****** initialize screen ******/
	initscr();
	getmaxyx(stdscr, yscreensize, xscreensize);

	if(xscreensize < 80 || yscreensize < 25){
		endwin();
		cout << "Error: Screen must be a minimum size of 80 X 25.\n";
		exit(1);
	}

	noecho();
	//raw();
	curs_set(0);
	clear();


	/********* initialize colours *********/
	start_color();
	init_pair(1, COLOR_BLACK, COLOR_WHITE);
	init_pair(2, COLOR_RED, COLOR_WHITE);
	init_pair(3, COLOR_WHITE, COLOR_BLACK);
	init_pair(4, COLOR_BLACK, COLOR_YELLOW);
	init_pair(CARDBACK_COLOUR, COLOR_GREEN, COLOR_BLUE);
	init_pair(PLAYER_TEXT_COLOUR, COLOR_GREEN, COLOR_BLACK);
	init_pair(PLAYER_TEXT_COLOUR + 1, COLOR_YELLOW, COLOR_BLACK);
	init_pair(MESSAGE_TEXT_COLOUR, COLOR_CYAN, COLOR_BLACK);
	numcolourpairs = 8;
	//bkgd(COLOR_PAIR(3));

	/** load the board **/
	if(load_board()) exit(2);

	/** initialize any needed windows **/
	initialize_windows();

/********** play the game *************/

	get_player_names();
	play_again = 100;
	do{
		mvprintw(0, 0, "%i ", play_again);
		refresh();
		/******* initialize values that need to be reset for every game ********/
		for(n = 0; n < numplayers; n++){
			player[n].score = player[n].priorscore = 0;
		}
		numrounds = game_over = 0;
		if(main_displaymode != none){
			clear();
			board.draw(boardy, boardx);
		}

		while(!game_over){
			if(main_displaymode != none){
				box(0, 0, yscreensize, boardx - 1, ' ');
				draw_cardback((yscreensize >> 1) - 1, 0);
			}

			numrounds ++;
			display_round_start(numrounds);
			deck.shuffle();
			deal_round();
			divide_hands();
			cut_deck();
			if(!game_over) play_hands();
			if(!game_over) peg_hands(main_displaymode);
	
			restock_deck();
	
		       	if(displaymode == ansi && !game_over){
				wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
				wprintw(messagewindow, "Press a key for the next round.\n");
				wrefresh(messagewindow);
				wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
				wait_for_keypress();
			}
			current_dealer = (current_dealer + 1) % numplayers;
		}

		if(displaymode == ansi) refresh();
		show_winner(displaymode);
		if(main_displaymode == ansi){
			wprintw(messagewindow, "Play again? (y/n)\n");
			wrefresh(messagewindow);
		}
/**
		do{
			c = wgetch(messagewindow);
			if(c >= 'A' && c <= 'Z') c -= ('A' - 'a');
		}while(c != 'y' && c != 'n');


		play_again = (c == 'y');
**/
		play_again --;
		current_dealer = (winning_player + 1) % numplayers;
		tally[winning_player] ++;
	}while(play_again);
	wprintw(messagewindow, "\nplayer 0: %i\nplayer 1:%i\n", tally[0], tally[1]);
	wrefresh(messagewindow);
	while(wgetch(messagewindow) != (char)27);
	
	delwin(messagewindow);
	endwin();

	return returnval;
}

void wait_for_keypress(void){
//	wgetch(messagewindow);
}

void draw_pegs(void){
	int n, x, y;

	if(main_displaymode == ansi){
		for(n = 0; n < numplayers; n++){
			if(player[n].score > 0){
				attron(COLOR_PAIR(board.pegcolour[n]) | A_BOLD);
				if(player[n].score < board.numholes + 1){

					x = board.hole[n][player[n].score - 1].x;
					y = board.hole[n][player[n].score - 1].y;
					move(y + boardy, x + boardx);
					addch('!');

					if(player[n].priorscore > 0){
						x = board.hole[n][player[n].priorscore - 1].x;
						y = board.hole[n][player[n].priorscore - 1].y;
						move(y + boardy, x + boardx);
						addch('!');
					}
				}else{
					x = board.endhole.x;
					y = board.endhole.y;
					move(y + boardy, x + boardx);
					addch('!');

					x = board.hole[n][player[n].priorscore - 1].x;
					y = board.hole[n][player[n].priorscore - 1].y;
					move(y + boardy, x + boardx);
					addch('!');
				}
				attroff(COLOR_PAIR(board.pegcolour[n]) | A_BOLD);
			}
			if(main_displaymode == ansi) wrefresh(messagewindow);
		}
	}
}

void increment_score(int playernum, int amount){
	/***** increment the score and assign it's old value to "priorscore" ******/
	player[playernum].priorscore = player[playernum].score;
	player[playernum].score += amount;
}

void restock_deck(void){
	int n;
	for(n = 0; n < numplayers; n++){
		deck += player[n].hand;
		player[n].hand.erase();
	}
	deck += crib;
	crib.erase();
	deck += cutcard;
	cutcard.erase();
}

void display_round_start(int numrounds){
	if(main_displaymode == ansi){
		wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD);
		wprintw(messagewindow, "\n=== Round %i =============\n", numrounds);
		wprintw(messagewindow, "%s deals.\n", player[current_dealer].name);
		wrefresh(messagewindow);
		wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD);
	}
}

void get_player_names(void){
	int n, m;
	for(n = 0; n < numplayers; n++){
		if(player[n].typeofplayer == 1){
			do{
				player[n].getname();
				for(m = 0; m < n && strcmp(player[n].name, player[m].name); m++);
			}while(m < n);
		}else{
			move(0, 0);
			curs_set(1);
			printw("Hi there player %i, what is your name? ", n + 1);
			echo();
			scanw("%s", player[n].name);
			noecho();
			curs_set(0);
			clear();
		}
	}
}

void show_winner(displaymodetype displaymode){
	if(displaymode != none){
		wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD | A_BLINK);
		wprintw(messagewindow, "%s wins!\n",  player[winning_player].name);
		wrefresh(messagewindow);
		wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD | A_BLINK);
	}

}

void draw_cardback(int y, int x){
	if(main_displaymode == ansi){
		move(y, x);
		attron(COLOR_PAIR(CARDBACK_COLOUR));
		printw("\\|/");
		move(y + 1, x);
		printw("-X-");
		move(y + 2, x);
		printw("/|\\");
		attroff(COLOR_PAIR(CARDBACK_COLOUR));	
	}
}

int initialize_table(int argc, char *argv[]){
	int n, returnval = 0, temp;
	/*** set the default values ***/
	numplayers = 2;					/* The number of players */
	handsize = 4;					/* The size of each player's hand after putting cards in the crib */
	starting_handsize = 6;				/* The number of cards initially dealt to each player */
	starting_cribsize = 0;				/* The number of cards initially dealt into the crib */
	current_dealer = rand() % numplayers;		/* The index number of the player who's dealing */
	add_to_crib_size = 2;				/* The number of cards each player needs to throw in the crib */
	maximum_score = 121;				/* The number of points required to win the game */
	game_over = 0;					/* non-zero if the game has been won */
	main_displaymode = none;			/* primary mode of display */

	if(argc > 1){
		for(n = 1; n < argc; n++){
			if(!strcmp(argv[n], "-f")){
				if(argc > n + 1){
					maximum_score = atoi(argv[n + 1]);
					n++;
				}else{
					returnval = 1;
					break;
				}
			}else if(!strcmp(argv[n], "-p")){
				if(argc > n + 1){
					temp = atoi(argv[n + 1]);
					switch(temp){
					    case 2:
						starting_handsize = 6;
						add_to_crib_size = 2;
						starting_cribsize = 0;
						numplayers = temp;
						break;
					    case 3:
						starting_handsize = 5;
						add_to_crib_size = 1;
						starting_cribsize = 1;
						numplayers = temp;
						break;
					    case 4:
						starting_handsize = 5;
						add_to_crib_size = 1;
						starting_cribsize = 0;
						numplayers = temp;
						break;
					}
					n++;
				}else{
					returnval = 1;
					break;
				}
			}else if(!strcmp(argv[n], "-s")){
				if(argc > n + 1){
					current_dealer = atoi(argv[n + 1]);
					n++;
				}else{
					returnval = 1;
					break;
				}
			}else if(!strcmp(argv[n], "-d")){
				if(argc > n + 1){
					n++;
					if(!strcmp(argv[n], "none")) main_displaymode = none;
					if(!strcmp(argv[n], "ansi")) main_displaymode = ansi;
				}else{
					returnval = 1;
					break;
				}
			}
		}
	}

	if(returnval == 1){
		cout << "Command Syntax:\n";
		cout << "cribbage [-f (finishing score)]\n";
		cout << "         [-p (number of players)]\n";
		cout << "         [-s (starting dealer)]\n";
		cout << "         [-d (display mode)]\n\n";
		cout << "\"display mode\" can be one of the following:\n";
		cout << "\tnone: Will not display anything.\n";
		cout << "\tansi: Displays the entire game in ansi highlighted text.\n";
	}

	return returnval;
}

int load_board(void){
	int returnval = 0, n, x, y;

	if(board.load("default.board")){
		returnval = 2;
	}else{

		baseboardcolour = numcolourpairs + 1;
		for(n = 0; n < board.numcolourpairs; n++){
			numcolourpairs++;
			init_pair(numcolourpairs, board.colourpair[n][0], board.colourpair[n][1]);
		}

		for(y = 0; y < board.ysize; y++){
			for(x = 0; x < board.xsize; x++){
				board.colour[y][x] += baseboardcolour;
			}
		}

		for(n = 0; n < board.numplayers; n++){
			board.pegcolour[n] += baseboardcolour;
		}

		boardy = 1;//(yscreensize - board.ysize) >> 1;
		boardx = xscreensize - board.xsize - 1;
	}

	return returnval;
}

void initialize_windows(void){
	int height, width;
	messagey = boardy + board.ysize;
	messagex = boardx;

	height = yscreensize - messagey - 2;
	width = board.xsize;

	messagewindow = newwin(height, width, messagey, messagex);
	scrollok(messagewindow, 1);	
}

void redraw_screen(void){
	if(main_displaymode != none){
		board.draw(boardy, boardx);
		draw_pegs();
		refresh();
		wrefresh(messagewindow);
	}
}

void peg_hands(displaymodetype displaymode){
	int n, m;
	for(n = 0; n < numplayers; n++){

		m = (n + current_dealer + 1) % numplayers;
		if(displaymode != none){
			wattron(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + m) | A_BOLD);
			wprintw(messagewindow, "\n%s pegs:\n", player[m].name);
			wattroff(messagewindow, A_BOLD);
		}
		increment_score(m, count_points(player[m].hand, HAND, displaymode));
		if(displaymode != none){
			redraw_screen();
			wattroff(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + m));
			wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
			wprintw(messagewindow, "Press any key to continue.\n\n");
			wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
			wrefresh(messagewindow);
			wait_for_keypress();
		}

		if(player[m].score >= maximum_score){
			game_over = 1;
			winning_player = m;
			break;
		}
	}

	if(!game_over){
		if(displaymode != none){
			wattron(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + current_dealer) | A_BOLD);
			wprintw(messagewindow, "And in the crib:\n");
			wattroff(messagewindow, A_BOLD);
			if(current_dealer == 0){
				crib.display(yscreensize - 5, 0);
			}else{
				crib.display(2, 0);
			}
			redraw_screen();
		}

		m = count_points(crib, CRIB, displaymode);
		if(m){
			increment_score(current_dealer, m);
			redraw_screen();
		}

		if(player[current_dealer].score >= maximum_score){
			game_over = 1;
			winning_player = current_dealer;
		}
	}

	redraw_screen();
}

int count_play_value(handtype pile, int current_player, displaymodetype displaymode){
	int n, m, sum, returnval = 0, testsize, runsize;
	handtype temphand;

	/******* check for fifteens or thirty-ones *******/
	sum = pile.sum();
	if(sum == 15){
		returnval += 2;
		if(displaymode != none) wprintw(messagewindow, "Fifteen two");
	}else if(sum == 31){ 
		returnval += 2;
		if(displaymode != none){
			wprintw(messagewindow, "Thirty-one for two");
		}

	}else if(pile.numcards == numplayers * handsize){
		returnval++;
		if(displaymode != none){
			wprintw(messagewindow, "Go for one");
		}
	}else{
		if(displaymode != none)
			wprintw(messagewindow, "%i", sum);
	}

	/******* check for pairs, pair royals, etc. ********/
	m = 1;
	for(n = pile.numcards - 1; n > 0 && pile.card[n].value == pile.card[n - 1].value; n--) m++;
	if(m >= 2){
		returnval += (m * (m - 1));

		if(displaymode != none){
			if(m == 2)
				wprintw(messagewindow, ", pair for two");
			else if(m == 3)
				wprintw(messagewindow, ", pair royal for %i",  m * (m - 1));
			else
				wprintw(messagewindow, ", double royal for %i",  m * (m - 1));
		}
	}

	/******** check for runs *******/
	if(pile.numcards > 2){
		runsize = 0;
		temphand.erase();

		temphand += pile.card[pile.numcards - 1];
		temphand += pile.card[pile.numcards - 2];

		for(testsize = 3; testsize <= pile.numcards; testsize ++){
			temphand += pile.card[pile.numcards - testsize];
			temphand.sort();
			m = 1;
			for(n = 1; n < testsize; n++){
				if(temphand.card[n].value - temphand.card[n - 1].value == 1){
					m++;
				}else{
					break;
				}
			}
			if(n >= testsize){
				runsize = m;
			}
		}
		if(runsize > 2){
			returnval += runsize;
			if(displaymode != none)
				wprintw(messagewindow, ", Run for %i", runsize);
		}
	}
	if(displaymode != none){
		wprintw(messagewindow, ".\n");
		wrefresh(messagewindow);
	}
	return returnval;
}

void play_hands(void){
	int x, y;
	int cardsplayed[numplayers], passedonce[numplayers];
	int total_cards_played = 0, current_numcards = 0, n;
	int current_player, last_player;
	handtype pile, original_hand[numplayers];

	for(n = 0; n < numplayers; n++){
		original_hand[n] = player[n].hand;
		cardsplayed[n] = passedonce[n] = 0;
	}

	pile.erase();
	current_player = (current_dealer + 1) % numplayers;

	while(total_cards_played < numplayers * handsize && !game_over){

		wattron(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + current_player) | A_BOLD);
		if((passedonce[current_player] != 1) && 
		   (player[current_player].play_card(&pile))){

			if(main_displaymode == ansi){
				x = (cardsplayed[current_player] << 2);
				if(current_player == 0) y = (yscreensize >> 1) + 3;
				else y = (yscreensize >> 1) - 5;
				cardsplayed[current_player]++;
				pile.card[current_numcards].display(y, x);

				if(current_player == 0) y = yscreensize - 3;
				else y = 1;
				x = (handsize - cardsplayed[current_player]) << 2;
				box(y, x, 3, 3, ' ');
			}

			current_numcards++;
			total_cards_played ++;

			n = count_play_value(pile, current_player, main_displaymode);
			if(n != 0){
				increment_score(current_player, n);
				redraw_screen();
			}

			if(player[current_player].score >= maximum_score){
				game_over = 1;
				winning_player = current_player;
				redraw_screen();

			}

			last_player = current_player;
		}else{

			if(last_player == current_player){
				pile.erase();
				current_numcards = 0;
				pile.numcards = 0;
				for(n = 0; n < numplayers; n++)
					passedonce[n] = 0;

				increment_score(current_player, 1);
				redraw_screen();

				if(player[current_player].score >= maximum_score){
					game_over = 1;
					winning_player = current_player;
				}

				if(main_displaymode == ansi){
					wprintw(messagewindow, "Go for one\n");
					wrefresh(messagewindow);
				}
			}else if(!passedonce[current_player]){
				if(main_displaymode == ansi){
					wprintw(messagewindow, "Pass\n");
					wrefresh(messagewindow);
				}
				passedonce[current_player] = 1;

			}
		}
		wattroff(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + current_player) | A_BOLD);

		if(pile.sum() == 31){
			pile.erase();
			current_numcards = 0;
			for(n = 0; n < numplayers; n++)
				passedonce[n] = 0;
		}

		current_player = (current_player + 1) % numplayers;
		if(main_displaymode != none){
			wrefresh(messagewindow);
			refresh();
		}

	}

	if(pile.numcards && !game_over){
		increment_score(last_player, 1);

		if(main_displaymode != none){
			wattron(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + last_player) | A_BOLD);
			wprintw(messagewindow, "One for last card.\n");
			wattroff(messagewindow, COLOR_PAIR(PLAYER_TEXT_COLOUR + last_player) | A_BOLD);
		}
	}

	if(main_displaymode != none){
		/**** erase the last card showing in the human player's hand and the text below ***/
		box(yscreensize - 6, 0, 5, 20, ' ');
		
		/*** update the board image ***/
		board.draw(boardy, boardx);
		draw_pegs();
		refresh();

		wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
		wprintw(messagewindow, "Press any key to continue.\n\n");
		wrefresh(messagewindow);
		wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));

		wait_for_keypress();

	}
	for(n = 0; n < numplayers; n++){
		player[n].hand = original_hand[n];
	}

	redraw_screen();

}

void deal_round(void){
	int n, m, l;

	/*** now deal the cards ***/
	for(n = 0; n < starting_handsize; n++){
		for(m = 0; m < numplayers; m++){
			l = (m + current_dealer) % numplayers;
			deck.deal(&player[l], 1);
		}
	}

	for(n = 0; n < starting_cribsize; n++){
		deck.numcards --;
		crib += deck.card[deck.numcards];
		deck.card[deck.numcards].erase();
	}
}

void cut_deck(void){
	int range, cutnum, minimum, x, y;
	minimum = 4;
	range = deck.numcards - 8;
	cutnum = minimum + rand() % range;
	cutcard = deck.draw(minimum + rand() % range);
	
	if(main_displaymode == ansi){
		x = 4;
		y = (yscreensize >> 1) - 2;
		mvprintw(y, x, "Cut:");
		x = 5;
		y = (yscreensize >> 1) - 1;
		cutcard.display(y, x);
	}

	if(cutcard.value == JACK){
		increment_score(current_dealer, 1);
		if(player[current_dealer].score >= maximum_score){
			game_over = 1;
			winning_player = current_dealer;
		}
		if(main_displaymode == ansi){
			x = 0;
			y = (yscreensize >> 1) + 3;
			wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD);
			wprintw(messagewindow, "Dealer gets one for Nobs.\n");
			wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD);
			wrefresh(messagewindow);
			draw_pegs();
		}
	}
	redraw_screen();
}

void randomize(void){
	srand(time(0));
}

void divide_hands(void){
	int n, x;
	
	for(n = 0; n < starting_handsize; n++){
		draw_cardback(1, n << 2);
	}

	for(n = 0; n < numplayers; n++){
		player[n].divide_hand();
	}

	x = (starting_handsize - add_to_crib_size) * 4;
	box(1, x, 3, add_to_crib_size << 2, ' ');
}

void box(int y1, int x1, int ysize, int xsize, char value){
	int y, y2;
	char *string;

	y2 = y1 + ysize - 1;
	string = new char[xsize + 1];
	memset((void *)string, value, xsize);
	string[xsize] = (char)NULL;
	
	for(y = y1; y <= y2; y++){
		mvprintw(y, x1, string);
	}

	delete string;

}

int count_fifteens(handtype hand, playmodetype playmode, displaymodetype displaymode){
	int score, numcombos, tally, n, m;
	handtype temphand;

	temphand = hand;
	if(playmode == HAND || playmode == CRIB) temphand += cutcard;
	temphand.sort();

	/************ looking for fifteens. ***********/
	numcombos = 1 << temphand.numcards;   /* This is the total number of combinations of the cards */
	score = 0;

	for(n = 1; n < numcombos; n++){
		tally = 0;
		for(m = 0; m < temphand.numcards; m++){
			if((1 << m) & n){
				tally += temphand.card[m].pointvalue;
			}
			
		}

		if(tally == 15){

			if(score != 0 && displaymode != none)
				wprintw(messagewindow, ", ");
			score += 2;
			if(displaymode != none) 
				wprintw(messagewindow, "Fifteen %i", score);
		}
	}
	if(displaymode != none && score != 0){
		wprintw(messagewindow, "\n");
		wrefresh(messagewindow);
	}

	return score;
}

int count_pairs(handtype hand, playmodetype playmode, displaymodetype displaymode){
	int n, m, alreadycounted, tally, score = 0, numpairs = 0;
	handtype temphand;

	temphand = hand;
	if(playmode == HAND || playmode == CRIB) temphand += cutcard;


	/********** looking for pairs and pair royals **********/
	for(n = 0; n < temphand.numcards; n++){
		/*** determine if the current card has already been counted in a pair combo ***/
		alreadycounted = 0;
		for(m = 0; m < n; m++){
			if(temphand.card[m].value == temphand.card[n].value)
				alreadycounted = 1;
		}

		if(!alreadycounted){
			tally = 1;
			for(m = n + 1; m < temphand.numcards; m++){
				if(temphand.card[n].value == temphand.card[m].value)
					tally++;
			}

			if(displaymode != none && tally > 1){
				switch(tally){
				    case 2: numpairs++; break;
				    case 3:
					if(score != 0) wprintw(messagewindow, ", ");
					wprintw(messagewindow, "Pair royal for six");
					break;
				    case 4: 
					if(score != 0) wprintw(messagewindow, ", ");
					wprintw(messagewindow, "Double royal for twelve");
					break;
				}
			}

			/** no if here, since tally - 1 == 0 if there are no pairs **/
			score += tally * (tally - 1);
		}
	}

	if(displaymode != none && score != 0){
		switch(numpairs){
		    case 0: break;
		    case 1:
			if(score > (numpairs << 1)) 
				wprintw(messagewindow, ", and a pair for two.");
			else
				wprintw(messagewindow, "Pair for two");
			break;
		    case 2:
			if(score > (numpairs << 1)) wprintw(messagewindow, ", ");
			if(rand() % 2)
				wprintw(messagewindow, "Pair of pairs for four");
			else
				wprintw(messagewindow, "Two pairs for four");
			break;
		    default:
			if(score > (numpairs << 1)) wprintw(messagewindow, ", ");
			wprintw(messagewindow, "%i pairs for %i", numpairs, numpairs << 1); break;


		}
		wprintw(messagewindow, "\n");
		wrefresh(messagewindow);
	}
	
	return score;
}

int count_runs(handtype hand, playmodetype playmode, displaymodetype displaymode){
	int n, runlength, multiplicand, newfactor, score, increment;
	handtype temphand;

	score = 0;
	runlength = multiplicand = newfactor = 1;

	temphand = hand;
	if(playmode == HAND || playmode == CRIB) temphand += cutcard;
	temphand.sort();

	for(n = 1; n < temphand.numcards; n++){
		if(temphand.card[n - 1].value == temphand.card[n].value){
			/*** if the current card is the same number as the previous***/
			newfactor ++;
		}else{
			if(newfactor > 1){
				multiplicand *= newfactor;
				newfactor = 1;
			}

			if(temphand.card[n - 1].value == temphand.card[n].value - 1){
				/*** if the curent card is one up from the previous **/
				runlength++;
			}else{
				if(runlength > 2){
					increment = runlength;
					increment *= multiplicand;
					score += increment;

					if(displaymode != none){
						switch(multiplicand){
						    case 2: wprintw(messagewindow, "Double run "); break;
						    case 3: wprintw(messagewindow, "Triple run "); break;
						    case 4: wprintw(messagewindow, "Quadruple run "); break;
						    default: wprintw(messagewindow, "Run ");
						}
						wprintw(messagewindow, "for %i", increment);
					}
				}
				multiplicand = newfactor = runlength = 1;
			}
		}
	}

	/***
	This repetition of the "if" needs to be done in case the loop
	ends with a run at the end of the hand.
	***/
	if(runlength > 2){
		if(newfactor > 1) multiplicand *= newfactor;
		increment = runlength;
		increment *= multiplicand;
		score += increment;

		if(displaymode != none){
			switch(multiplicand){
			    case 2: wprintw(messagewindow, "Double run "); break;
			    case 3: wprintw(messagewindow, "Triple run "); break;
			    case 4: wprintw(messagewindow, "Quadruple run "); break;
			    default: wprintw(messagewindow, "Run ");
			}
			wprintw(messagewindow, "for %i", increment);
		}
	}

	if(score > 0 && displaymode != none){
		wprintw(messagewindow, "\n");
		wrefresh(messagewindow);
	}

	return score;
}

int count_flush(handtype hand, playmodetype playmode, displaymodetype displaymode){
	int score = 0, n, isaflush = 1;

	for(n = 1; n < hand.numcards; n++){
		if(hand.card[n].suit != hand.card[n - 1].suit)
			isaflush = 0;
	}

	if(isaflush){
		if(playmode == CRIB && cutcard.suit != hand.card[0].suit) isaflush = 0;
	}

	if(isaflush){
		score = hand.numcards;
		if(cutcard.suit == hand.card[0].suit) score++;
	}

	if(score != 0  && displaymode != none){
		wprintw(messagewindow, "%i card flush\n", score);
		wrefresh(messagewindow);
	}
	return score;
}

int count_nobs(handtype hand, playmodetype playmode, displaymodetype displaymode){
	int returnval = 0, n;

	if(cutcard.value != JACK && playmode != PEG){
		for(n = 0; n < hand.numcards; n++){
			if(hand.card[n].value == JACK
			   && hand.card[n].suit == cutcard.suit)
				returnval = 1;
		}
	}

	if(displaymode != none && returnval != 0){
		wprintw(messagewindow, "1 for his nobs\n");
		wrefresh(messagewindow);
	}

	return returnval;
}

int count_points(handtype hand, playmodetype playmode, displaymodetype displaymode){
	/***
	This goes through every possible type of score and checks
	to see if any of them can be found in the hand.  eg.  first
	it looks for anything that adds up to fifteen, then it looks
	for runs, and so on.  The individual checks are performed
	by separate functions.
	***/
	int score;

	score = count_fifteens(hand, playmode, displaymode);
	score += count_runs(hand, playmode, displaymode);
	score += count_pairs(hand, playmode, displaymode);
	score += count_nobs(hand, playmode, displaymode);
	score += count_flush(hand, playmode, displaymode);	

	/*** display the score ***/
	if(displaymode != none){
		if(score == 0)
			switch(rand() % 10){
				case 0: wprintw(messagewindow, "Good old \"Nineteen\".\n"); break;
				case 1: wprintw(messagewindow, "Nothing whatsoever.\n"); break;
				case 2: wprintw(messagewindow, "Zero, Zip and Zilch.\n"); break;
				case 3: wprintw(messagewindow, "Absolutely nothing.\n"); break;
				default: wprintw(messagewindow, "Nothing at all.\n");
			}

		else{
			wprintw(messagewindow, "for a total of %i.\n", score);
		}
		wrefresh(messagewindow);
	}

	return score;
}



/*** 
This returns a 1 or 0 depending on whether or not it can play a card
without going over 31, and if so, it puts that card into the hand "pile".
it does NOT add to the score.  This is done elsewhere.
***/
int playertype::play_card(handtype *pile){
	int returnval = 0, n, bestscore = -1, newscore, bestplaycard = -1, tally, cardpicked;
	handtype testhand;
	int gotone;
	unsigned char keypressed;

	if(pile->numcards < MAXHANDSIZE){

		/*******************************************
		This part of the if handles User I/O if the
		player is a human one.
		*******************************************/
		if(typeofplayer == 0){
			if(hand.numcards > 0){

				/** first, display the cards and options for the player to see **/
				box(yscreensize - 6, 0, 4, (starting_handsize << 2), ' ');
				wprintw(messagewindow, "%s: ", name);

				for(n = 0; n < hand.numcards; n++){
					hand.card[n].display(yscreensize - 6, n * 4);
					move(yscreensize - 3, n * 4);
					printw("[%c] ", (char)((int)'a' + n));
				}
				printw("\nHit space to pass");
				refresh();
				wrefresh(messagewindow);

				gotone = 0;
				do{

					keypressed = getch();
					if(keypressed >= 'A' && keypressed <= 'Z') keypressed += (int)'a' - (int)'A';
					if(keypressed >= 'a' && keypressed < (char)((int)'a' + hand.numcards)) gotone = 1;
					if(keypressed == ' ') gotone = 1;

					/******* now check to see if it's a valid option ******/
					if(gotone){
						cardpicked = (int)(keypressed - 'a');
						tally = pile->sum();
						if(keypressed == ' '){
							/**** make sure they actually can't play a card ****/
							for(n = 0; n < hand.numcards && (tally + hand.card[n].pointvalue > 31); n++);
							if(n < hand.numcards) gotone = 0;
						}else{
							/**** make sure the selected card can actually be played ****/
							newscore = tally + hand.card[cardpicked].pointvalue;
							if(newscore > 31) gotone = 0;
						}
					}
				}while(!gotone);

				if(keypressed != ' '){
					returnval = 1;
					pile->card[pile->numcards] = hand.card[cardpicked];
					pile->numcards ++;
					hand.removecard(cardpicked);
					returnval = 1;
				}
			}

		/*******************************************
		This part of the if handles the computer player.
		*******************************************/
		}else{
			if(main_displaymode == ansi){
				wprintw(messagewindow, "%s: ", name);
				wrefresh(messagewindow);
			}
			tally = pile->sum();
			for(n = 0; n < hand.numcards; n++){
				if(tally + hand.card[n].pointvalue <= 31){
					testhand = *pile;
					testhand += hand.card[n];
					newscore = count_play_value(testhand, 0, none);
					if(newscore > bestscore){
						bestplaycard = n;
						bestscore = newscore;
					}
				}
			}

			if(bestscore == 0){
				do{
					n = rand() % hand.numcards;
				}while(tally + hand.card[n].pointvalue > 31);
				bestplaycard = n;
			}

			if(bestplaycard != -1){
				pile->card[pile->numcards] = hand.card[bestplaycard];
				pile->numcards++;
				hand.removecard(bestplaycard);
				returnval = 1;
			}
		}
	}else{
		wprintw(messagewindow, "\n<<<<<< ERROR: Not enough space in the pile >>>>>>\n");
	}
	return returnval;
}


void playertype::divide_by_human(void){
	unsigned char keypressed;
	int n, m, done = 0;
	displaymodetype displaymode = main_displaymode;
	int numcardspicked = 0, card_picked[MAXHANDSIZE];

	if(add_to_crib_size){
		hand.sort();

		/** first, display the cards and options for the player to see **/
		if(displaymode != none){

			for(n = 0; n < hand.numcards; n++){
				hand.card[n].display(yscreensize - 6, n * 4);
				move(yscreensize - 3, n * 4);
				printw("[%c] ", (char)((int)'a' + n));
			}
			refresh();

			wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
			if(player[current_dealer].number != number)
				wprintw(messagewindow, "Select which cards to put into %s's crib.\n", player[current_dealer].name);
			else
				wprintw(messagewindow, "Select which cards to put into your crib.\n");

			wprintw(messagewindow, "(press the space bar when ready)\n");
			wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR));
			wrefresh(messagewindow);

		}

		/*** now let them pick out the cards they pass to the crib ***/
		numcardspicked = 0;
		do{
			keypressed = getch();

			if(keypressed == (char)27){
				endwin();
				exit(1);
			}

			if(keypressed == ' '){
				if(numcardspicked != add_to_crib_size){
					wattron(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD);
					if(numcardspicked < add_to_crib_size)
						wprintw(messagewindow, "You need to put %i cards in the crib.\n", add_to_crib_size);
					else
						wprintw(messagewindow, "You can only put %i cards in the crib.\n", add_to_crib_size);
					wattroff(messagewindow, COLOR_PAIR(MESSAGE_TEXT_COLOUR) | A_BOLD);
					wrefresh(messagewindow);
				}else{
					done = 1;
				}
			}

			for(n = 0; n < hand.numcards && (keypressed != (unsigned char)((int)'a' + n)); n++);
			if(n < hand.numcards){
				for(m = 0; m < numcardspicked; m++)
					if(card_picked[m] == n) break;

				if(m >= numcardspicked){
					card_picked[numcardspicked] = n;
					numcardspicked ++;

					attron(COLOR_PAIR(4));

					move(yscreensize - 3, 4 * n);
					printw("[%c]", 'A' + keypressed - 'a');
					attroff(COLOR_PAIR(4));

					hand.card[n].display(yscreensize - 6, n * 4);

				}else{
					move(yscreensize - 3, 4 * n);
					printw("[%c]", keypressed);

					numcardspicked--;
					for(; m < numcardspicked; m++){
						card_picked[m] = card_picked[m + 1];
					}					
				}
				move(0, 0);
			}
		}while(!done);

		/*** cards have been selected, now throw them in the crib ***/
		for(n = 0; n < numcardspicked; n++){
			hand.passcard(&crib, card_picked[n]);
			/** IMPORTANT
			this additional loop needs to be here, because
			the cards get rearranged in the hand when one is removed;
			**/

			for(m = n + 1; m < numcardspicked; m++)
				if(card_picked[m] > card_picked[n]) card_picked[m]--;
		}
	}
}

void playertype::divide_by_computer(void){
	int combo, max, n, m, numbits, bestcombo, bestscore = -1, score;
	int potentialscore, bestpotentialscore, bestpotentialcombo;
	int a, b;
	handtype testhand;
	decktype testdeck;

	max = 1 << hand.numcards;

	/******
	First, calculate how to get the most points without
	taking whatever card is cut into account.
	******/
	for(combo = 0; combo < max; combo++){
		numbits = 0;
		for(n = 0; n < hand.numcards; n++){
			if(combo & (1 << n)) numbits++;
		}

		if(numbits == add_to_crib_size){
			/***Here is where we check the value of the hand ***/
			testhand.erase();

			/** generate the new hand **/
			for(n = 0; n < hand.numcards; n++){
				if(!(combo & (1 << n))){
					testhand += hand.card[n];
				}
			}

			/** get the value of it **/
			score = count_points(testhand, NOCUT, none);
			if(score > bestscore ){
				bestscore = score;
				bestcombo = combo;
			}
		}
	}

	if(typeofplayer == 2){
		/*******
		Now, take a look at every combination and see which one
		offers the most ~potential~ points.
		*******/
		for(combo = 0; combo < max; combo++){
			numbits = 0;
			for(n = 0; n < hand.numcards; n++){
				if(combo & (1 << n)) numbits++;
			}

			if(numbits == add_to_crib_size){
				for(n = 0; n < 52; n++){
					testhand.erase();

					/** generate the new potential hand **/
					for(m = 0; m < hand.numcards; m++){
						if(!(combo & (1 << m))){
							testhand += hand.card[m];
						}
					}


					for(m = 0; m < testhand.numcards && (testhand.card[m] != testdeck.card[n]); m++);
					if(m == testhand.numcards){
						testhand += testdeck.card[n];

						/** get the value of it **/
						potentialscore = count_points(testhand, NOCUT, none);
						if(potentialscore > bestpotentialscore ){
							bestpotentialscore = potentialscore;
							bestpotentialcombo = combo;
						}
					}
				}
			}
		}

		/**************************************************************
		We have now calculated the cut that would give us the best
		definite number of points, as well at the cut that will
		give us the best *potential* number of points.  Now give them
		each a certain weight and decide which cut to use.
		Note that this does not yet account for the probablility of
		actually getting the potential hand.
		**************************************************************/

		//if(bestscore < bestpotentialscore && bestscore < 8 && bestcombo != bestpotentialcombo)

		a = bestscore;
		b = bestpotentialscore;
		if(b > (a * a) >> 2)
		{
			bestcombo = bestpotentialcombo;
			wprintw(messagewindow, "\n***** Taking Potential Hand*****\n");
		}else{
			wprintw(messagewindow, "\n===== Taking given hand =====\n");
		}
		wrefresh(messagewindow);
	}

	/***********************
	very important:  This loop MUST go in this order because the hand
	moves all the cards after the one pulled to fill in the space.  If
	They are grabbed in ascending order, everything after the first one
	will be wrong.
	***********************/
	for(n = hand.numcards - 1; n >= 0; n--)
		if((1 << n) & bestcombo)
			hand.passcard(&crib, n);
}

